from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Dict
from typing import List

import xarray as xr
from typing_inspect import get_args
from typing_inspect import is_generic_type

from glotaran.model.item import model_item
from glotaran.model.item import model_item_typed
from glotaran.plugin_system.megacomplex_registration import register_megacomplex

if TYPE_CHECKING:
    from typing import Any

    from glotaran.model import DatasetModel


def create_model_megacomplex_type(
    megacomplex_types: dict[str, Megacomplex], default_type: str = None
) -> type:
    @model_item_typed(types=megacomplex_types, default_type=default_type)
    class ModelMegacomplex:
        """This class holds all Megacomplex types defined by a model."""

    return ModelMegacomplex


def megacomplex(
    *,
    dimension: str | None = None,
    model_items: dict[str, dict[str, Any]] = None,
    properties: Any | dict[str, dict[str, Any]] = None,
    dataset_model_items: dict[str, dict[str, Any]] = None,
    dataset_properties: Any | dict[str, dict[str, Any]] = None,
    unique: bool = False,
    register_as: str | None = None,
):
    """The `@megacomplex` decorator is intended to be used on subclasses of
    :class:`glotaran.model.Megacomplex`. It registers the megacomplex model
    and makes it available in analysis models.
    """
    properties = properties if properties is not None else {}
    properties["dimension"] = {"type": str}
    if dimension is not None:
        properties["dimension"]["default"] = dimension

    if model_items is None:
        model_items = {}
    else:
        model_items, properties = _add_model_items_to_properties(model_items, properties)

    dataset_properties = dataset_properties if dataset_properties is not None else {}
    if dataset_model_items is None:
        dataset_model_items = {}
    else:
        dataset_model_items, dataset_properties = _add_model_items_to_properties(
            dataset_model_items, dataset_properties
        )

    def decorator(cls):

        setattr(cls, "_glotaran_megacomplex_model_items", model_items)
        setattr(cls, "_glotaran_megacomplex_dataset_model_items", dataset_model_items)
        setattr(cls, "_glotaran_megacomplex_dataset_properties", dataset_properties)
        setattr(cls, "_glotaran_megacomplex_unique", unique)

        megacomplex_type = model_item(properties=properties, has_type=True)(cls)

        if register_as is not None:
            megacomplex_type.name = register_as
            register_megacomplex(register_as, megacomplex_type)

        return megacomplex_type

    return decorator


def _add_model_items_to_properties(model_items: dict, properties: dict) -> tuple[dict, dict]:
    for name, item in model_items.items():
        item_type = item["type"] if isinstance(item, dict) else item
        property_type = str

        if is_generic_type(item_type):
            if item_type._name == "List":
                property_type = List[str]
                item_type = get_args(item_type)[0]
            elif item_type._name == "Dict":
                property_type = Dict[str, str]
                item_type = get_args(item_type)[1]

        property_dict = item.copy() if isinstance(item, dict) else {}
        property_dict["type"] = property_type
        properties[name] = property_dict
        model_items[name] = item_type
    return model_items, properties


class Megacomplex:
    """A base class for megacomplex models.

    Subclasses must overwrite :method:`glotaran.model.Megacomplex.calculate_matrix`
    and :method:`glotaran.model.Megacomplex.index_dependent`.
    """

    def calculate_matrix(
        self,
        dataset_model: DatasetModel,
        indices: dict[str, int],
        **kwargs,
    ) -> xr.DataArray:
        raise NotImplementedError

    def index_dependent(self, dataset_model: DatasetModel) -> bool:
        raise NotImplementedError

    def finalize_data(
        self,
        dataset_model: DatasetModel,
        dataset: xr.Dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        raise NotImplementedError

    @classmethod
    def glotaran_model_items(cls) -> str:
        return cls._glotaran_megacomplex_model_items

    @classmethod
    def glotaran_dataset_model_items(cls) -> str:
        return cls._glotaran_megacomplex_dataset_model_items

    @classmethod
    def glotaran_dataset_properties(cls) -> str:
        return cls._glotaran_megacomplex_dataset_properties

    @classmethod
    def glotaran_unique(cls) -> bool:
        return cls._glotaran_megacomplex_unique



from __future__ import annotations

from warnings import warn

import numpy as np
from scipy.optimize import OptimizeResult
from scipy.optimize import least_squares

from glotaran import __version__ as glotaran_version
from glotaran.analysis.optimization_group import OptimizationGroup
from glotaran.parameter import ParameterHistory
from glotaran.project import Result
from glotaran.project import Scheme

SUPPORTED_METHODS = {
    "TrustRegionReflection": "trf",
    "Dogbox": "dogbox",
    "Levenberg-Marquardt": "lm",
}


def optimize(scheme: Scheme, verbose: bool = True, raise_exception: bool = False) -> Result:

    optimization_groups = [
        OptimizationGroup(scheme, group) for group in scheme.model.get_dataset_groups().values()
    ]

    (
        free_parameter_labels,
        initial_parameter,
        lower_bounds,
        upper_bounds,
    ) = scheme.parameters.get_label_value_and_bounds_arrays(exclude_non_vary=True)

    if scheme.optimization_method not in SUPPORTED_METHODS:
        raise ValueError(
            f"Unsupported optimization method {scheme.optimization_method}. "
            f"Supported methods are '{list(SUPPORTED_METHODS.keys())}'"
        )
    method = SUPPORTED_METHODS[scheme.optimization_method]

    nfev = scheme.maximum_number_function_evaluations
    ftol = scheme.ftol
    gtol = scheme.gtol
    xtol = scheme.xtol
    verbose = 2 if verbose else 0
    termination_reason = ""

    parameter_history = ParameterHistory()
    parameter_history.append(scheme.parameters)
    try:
        ls_result = least_squares(
            _calculate_penalty,
            initial_parameter,
            bounds=(lower_bounds, upper_bounds),
            method=method,
            max_nfev=nfev,
            verbose=verbose,
            ftol=ftol,
            gtol=gtol,
            xtol=xtol,
            kwargs={
                "free_parameter_labels": free_parameter_labels,
                "optimization_groups": optimization_groups,
                "parameter_history": parameter_history,
            },
        )
        termination_reason = ls_result.message
    except Exception as e:
        if raise_exception:
            raise e
        warn(f"Optimization failed:\n\n{e}")
        termination_reason = str(e)
        ls_result = None

    return _create_result(
        scheme,
        optimization_groups,
        ls_result,
        free_parameter_labels,
        termination_reason,
        parameter_history,
    )


def _calculate_penalty(
    parameters: np.ndarray,
    *,
    free_parameter_labels: list[str],
    optimization_groups: list[OptimizationGroup],
    parameter_history: ParameterHistory,
):
    for group in optimization_groups:
        group.parameters.set_from_label_and_value_arrays(free_parameter_labels, parameters)
        group.reset()
    parameter_history.append(
        optimization_groups[0].parameters
    )  # parameters are the same for all groups

    penalties = [group.full_penalty for group in optimization_groups]

    return np.concatenate(penalties) if len(penalties) != 1 else penalties[0]


def _create_result(
    scheme: Scheme,
    optimization_groups: list[OptimizationGroup],
    ls_result: OptimizeResult | None,
    free_parameter_labels: list[str],
    termination_reason: str,
    parameter_history: ParameterHistory,
) -> Result:

    success = ls_result is not None

    number_of_function_evaluation = (
        ls_result.nfev if success else parameter_history.number_of_records
    )
    number_of_jacobian_evaluation = ls_result.njev if success else None
    optimality = float(ls_result.optimality) if success else None
    number_of_data_points = ls_result.fun.size if success else None
    number_of_variables = ls_result.x.size if success else None
    degrees_of_freedom = number_of_data_points - number_of_variables if success else None
    chi_square = float(np.sum(ls_result.fun ** 2)) if success else None
    reduced_chi_square = chi_square / degrees_of_freedom if success else None
    root_mean_square_error = float(np.sqrt(reduced_chi_square)) if success else None
    jacobian = ls_result.jac if success else None

    if success:
        for group in optimization_groups:
            group.parameters.set_from_label_and_value_arrays(free_parameter_labels, ls_result.x)
            group.reset()

    data = {}
    for group in optimization_groups:
        data.update(
            group.create_result_data(parameter_history, success=success, add_svd=scheme.add_svd)
        )

    # the optimized parameters are those of the last run if the optimization has crashed
    parameters = optimization_groups[0].parameters
    covariance_matrix = None
    if success:
        # See PR #706: More robust covariance matrix calculation
        _, jacobian_SV, jacobian_RSV = np.linalg.svd(jacobian, full_matrices=False)
        jacobian_SV_square = jacobian_SV ** 2
        mask = jacobian_SV_square > np.finfo(float).eps
        covariance_matrix = (jacobian_RSV[mask].T / jacobian_SV_square[mask]) @ jacobian_RSV[mask]
        standard_errors = root_mean_square_error * np.sqrt(np.diag(covariance_matrix))
        for label, error in zip(free_parameter_labels, standard_errors):
            parameters.get(label).standard_error = error

    additional_penalty = [group.additional_penalty for group in optimization_groups]

    cost = [group.cost for group in optimization_groups]

    return Result(
        additional_penalty=additional_penalty,
        cost=cost,
        data=data,
        glotaran_version=glotaran_version,
        free_parameter_labels=free_parameter_labels,
        number_of_function_evaluations=number_of_function_evaluation,
        initial_parameters=scheme.parameters,
        optimized_parameters=parameters,
        parameter_history=parameter_history,
        scheme=scheme,
        success=success,
        termination_reason=termination_reason,
        chi_square=chi_square,
        covariance_matrix=covariance_matrix,
        degrees_of_freedom=degrees_of_freedom,
        jacobian=jacobian,
        number_of_data_points=number_of_data_points,
        number_of_jacobian_evaluations=number_of_jacobian_evaluation,
        number_of_variables=number_of_variables,
        optimality=optimality,
        reduced_chi_square=reduced_chi_square,
        root_mean_square_error=root_mean_square_error,
    )



from __future__ import annotations

import importlib
from collections import namedtuple
from copy import deepcopy

import numpy as np

from glotaran.analysis.optimize import optimize
from glotaran.analysis.simulation import simulate
from glotaran.builtin.megacomplexes.decay import DecayMegacomplex
from glotaran.builtin.megacomplexes.spectral import SpectralMegacomplex
from glotaran.io import prepare_time_trace_dataset
from glotaran.model import Megacomplex
from glotaran.model import Model
from glotaran.parameter import ParameterGroup
from glotaran.project import Scheme

ParameterSpec = namedtuple("ParameterSpec", "base equal_area shapes")
NoiseSpec = namedtuple("NoiseSpec", "active seed std_dev")
SimulationSpec = namedtuple("SimulationSpec", "max_nfev noise")
DatasetSpec = namedtuple("DatasetSpec", "times wavelengths irf shapes")
IrfSpec = namedtuple("IrfSpec", "location width")
ShapeSpec = namedtuple("ShapeSpec", "amplitude location width")
ModelSpec = namedtuple("ModelSpec", "base shape spectral_megacomplex equ_area")
OptimizationSpec = namedtuple("OptimizationSpec", "nnls max_nfev")


class SpectralDecayModel(Model):
    @classmethod
    def from_dict(
        cls,
        model_dict,
        *,
        megacomplex_types: dict[str, type[Megacomplex]] | None = None,
        default_megacomplex_type: str | None = None,
    ):
        defaults: dict[str, type[Megacomplex]] = {
            "decay": DecayMegacomplex,
            "spectral": SpectralMegacomplex,
        }
        if megacomplex_types is not None:
            defaults.update(megacomplex_types)
        return super().from_dict(
            model_dict,
            megacomplex_types=defaults,
            default_megacomplex_type=default_megacomplex_type,
        )


def plot_overview(res, title=None):
    """very simple plot helper function derived from pyglotaran_extras"""
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots(2, 2, figsize=(9, 8), constrained_layout=True)
    if title:
        fig.suptitle(title, fontsize=16)
    sas = res.species_associated_spectra
    traces = res.species_concentration
    if "spectral" in traces.coords:
        traces.sel(spectral=res.spectral[0], method="nearest").plot.line(x="time", ax=ax[0, 0])
    else:
        traces.plot.line(x="time", ax=ax[0, 0])
    sas.plot.line(x="spectral", ax=ax[0, 1])
    rLSV = res.residual_left_singular_vectors
    rLSV.isel(left_singular_value_index=range(min(2, len(rLSV)))).plot.line(x="time", ax=ax[1, 0])

    ax[1, 0].set_title("res. LSV")
    rRSV = res.residual_right_singular_vectors
    rRSV.isel(right_singular_value_index=range(min(2, len(rRSV)))).plot.line(
        x="spectral", ax=ax[1, 1]
    )

    ax[1, 1].set_title("res. RSV")
    plt.show(block=False)


def test_equal_area_penalties(debug=False):
    # %%

    optim_spec = OptimizationSpec(nnls=True, max_nfev=999)
    noise_spec = NoiseSpec(active=True, seed=1, std_dev=1e-8)

    wavelengths = np.arange(650, 670, 2)
    time_p1 = np.linspace(-1, 2, 50, endpoint=False)
    time_p2 = np.linspace(2, 10, 30, endpoint=False)
    time_p3 = np.geomspace(10, 50, num=20)
    times = np.concatenate([time_p1, time_p2, time_p3])

    irf_loc = float(times[20])
    irf_width = float((times[1] - times[0]) * 10)
    irf = IrfSpec(irf_loc, irf_width)

    amplitude = 1
    location1 = float(wavelengths[2])  # 2
    location2 = float(wavelengths[-3])  # -3
    width1 = float((wavelengths[1] - wavelengths[0]) * 5)
    width2 = float((wavelengths[1] - wavelengths[0]) * 3)
    shape1 = ShapeSpec(amplitude, location1, width1)
    shape2 = ShapeSpec(amplitude, location2, width2)
    dataset_spec = DatasetSpec(times, wavelengths, irf, [shape1, shape2])

    wavelengths = dataset_spec.wavelengths
    equ_interval = [(min(wavelengths), max(wavelengths))]
    weight = 0.01
    # %% The base model specification (mspec)
    base = {
        "initial_concentration": {
            "j1": {
                "compartments": ["s1", "s2"],
                "parameters": ["i.1", "i.2"],
            },
        },
        "megacomplex": {
            "mc1": {"type": "decay", "k_matrix": ["k1"]},
        },
        "k_matrix": {
            "k1": {
                "matrix": {
                    ("s1", "s1"): "kinetic.1",
                    ("s2", "s2"): "kinetic.2",
                }
            }
        },
        "irf": {
            "irf1": {"type": "gaussian", "center": "irf.center", "width": "irf.width"},
        },
        "dataset": {
            "dataset1": {
                "initial_concentration": "j1",
                "megacomplex": ["mc1"],
                "irf": "irf1",
            },
        },
    }

    shape = {
        "shape": {
            "sh1": {
                "type": "gaussian",
                "amplitude": "shapes.amps.1",
                "location": "shapes.locs.1",
                "width": "shapes.width.1",
            },
            "sh2": {
                "type": "gaussian",
                "amplitude": "shapes.amps.2",
                "location": "shapes.locs.2",
                "width": "shapes.width.2",
            },
        }
    }

    spectral_megacomplex = {
        "type": "spectral",
        "shape": {
            "s1": "sh1",
            "s2": "sh2",
        },
    }

    equ_area = {
        "clp_area_penalties": [
            {
                "source": "s1",
                "target": "s2",
                "parameter": "rela.1",
                "source_intervals": equ_interval,
                "target_intervals": equ_interval,
                "weight": weight,
            },
        ],
    }
    mspec = ModelSpec(base, shape, spectral_megacomplex, equ_area)

    rela = 1.0  # relation between areas
    irf = dataset_spec.irf
    [sh1, sh2] = dataset_spec.shapes
    pspec_base = {
        "kinetic": [1e-1, 5e-3],
        "i": [0.5, 0.5, {"vary": False}],
        "irf": [["center", irf.location], ["width", irf.width]],
    }
    pspec_equa_area = {
        "rela": [rela, {"vary": False}],
    }
    pspec_shape = {
        "shapes": {
            "amps": [sh1.amplitude, sh2.amplitude],
            "locs": [sh1.location, sh2.location],
            "width": [sh1.width, sh2.width],
        },
    }
    pspec = ParameterSpec(pspec_base, pspec_equa_area, pspec_shape)

    # derivates:
    mspec_sim = dict(deepcopy(mspec.base), **mspec.shape)
    mspec_sim["megacomplex"]["mc2"] = mspec.spectral_megacomplex
    mspec_sim["dataset"]["dataset1"]["global_megacomplex"] = ["mc2"]

    mspec_fit_wp = dict(deepcopy(mspec.base), **mspec.equ_area)
    mspec_fit_np = dict(deepcopy(mspec.base))

    model_sim = SpectralDecayModel.from_dict(mspec_sim)
    model_wp = SpectralDecayModel.from_dict(mspec_fit_wp)
    model_np = SpectralDecayModel.from_dict(mspec_fit_np)
    print(model_np)

    # %% Parameter specification (pspec)

    pspec_sim = dict(deepcopy(pspec.base), **pspec.shapes)
    param_sim = ParameterGroup.from_dict(pspec_sim)

    # For the wp model we create two version of the parameter specification
    # One has all inputs fixed, the other has all but the first free
    # for both we perturb kinetic parameters a bit to give the optimizer some work
    pspec_wp = dict(deepcopy(pspec.base), **pspec.equal_area)
    pspec_wp["kinetic"] = [v * 1.01 for v in pspec_wp["kinetic"]]
    pspec_wp["i"] = [[1, {"vary": False}], 1]
    pspec_np = dict(deepcopy(pspec.base))

    param_wp = ParameterGroup.from_dict(pspec_wp)
    param_np = ParameterGroup.from_dict(pspec_np)

    # %% Print models with parameters
    print(model_sim.markdown(param_sim))
    print(model_wp.markdown(param_wp))
    print(model_np.markdown(param_np))

    # %%
    simulated_data = simulate(
        model_sim,
        "dataset1",
        param_sim,
        coordinates={"time": times, "spectral": wavelengths},
        noise=noise_spec.active,
        noise_std_dev=noise_spec.std_dev,
        noise_seed=noise_spec.seed,
    )
    # %%
    simulated_data = prepare_time_trace_dataset(simulated_data)
    # make a copy to keep an intact reference
    data = deepcopy(simulated_data)

    # %% Optimizing model without penalty (np)

    model_np.dataset_group_models["default"].method = (
        "non_negative_least_squares" if optim_spec.nnls else "variable_projection"
    )

    dataset = {"dataset1": data}
    scheme_np = Scheme(
        model=model_np,
        parameters=param_np,
        data=dataset,
        maximum_number_function_evaluations=optim_spec.max_nfev,
    )
    result_np = optimize(scheme_np)
    print(result_np)

    model_wp.dataset_group_models["default"].method = (
        "non_negative_least_squares" if optim_spec.nnls else "variable_projection"
    )

    # %% Optimizing model with penalty fixed inputs (wp_ifix)
    scheme_wp = Scheme(
        model=model_wp,
        parameters=param_wp,
        data=dataset,
        maximum_number_function_evaluations=optim_spec.max_nfev,
    )
    result_wp = optimize(scheme_wp)
    print(result_wp)

    if debug:
        # %% Plot results
        plt_spec = importlib.util.find_spec("matplotlib")
        if plt_spec is not None:
            import matplotlib.pyplot as plt

            plot_overview(result_np.data["dataset1"], "no penalties")
            plot_overview(result_wp.data["dataset1"], "with penalties")
            plt.show()

    # %% Test calculation
    print(result_wp.data["dataset1"])
    area1_np = np.sum(result_np.data["dataset1"].species_associated_spectra.sel(species="s1"))
    area2_np = np.sum(result_np.data["dataset1"].species_associated_spectra.sel(species="s2"))
    print("area_np", area1_np, area2_np)
    assert not np.isclose(area1_np, area2_np)

    area1_wp = np.sum(result_wp.data["dataset1"].species_associated_spectra.sel(species="s1"))
    area2_wp = np.sum(result_wp.data["dataset1"].species_associated_spectra.sel(species="s2"))
    assert np.isclose(area1_wp, area2_wp)

    input_ratio = result_wp.optimized_parameters.get("i.1") / result_wp.optimized_parameters.get(
        "i.2"
    )
    print("input", input_ratio)
    assert np.isclose(input_ratio, 1.5038858115)


if __name__ == "__main__":
    test_equal_area_penalties(debug=True)
    test_equal_area_penalties(debug=False)



"""Data Io registration convenience functions.

Note
----
The [call-arg] type error would be raised since the base methods doesn't have a ``**kwargs``
argument, but we rather ignore this error here, than adding ``**kwargs`` to the base method
and causing an [override] type error in the plugins implementation.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from tabulate import tabulate

from glotaran.io.interface import DataIoInterface
from glotaran.plugin_system.base_registry import __PluginRegistry
from glotaran.plugin_system.base_registry import add_instantiated_plugin_to_registry
from glotaran.plugin_system.base_registry import get_method_from_plugin
from glotaran.plugin_system.base_registry import get_plugin_from_registry
from glotaran.plugin_system.base_registry import is_registered_plugin
from glotaran.plugin_system.base_registry import methods_differ_from_baseclass_table
from glotaran.plugin_system.base_registry import registered_plugins
from glotaran.plugin_system.base_registry import set_plugin
from glotaran.plugin_system.base_registry import show_method_help
from glotaran.plugin_system.io_plugin_utils import bool_table_repr
from glotaran.plugin_system.io_plugin_utils import inferr_file_format
from glotaran.plugin_system.io_plugin_utils import not_implemented_to_value_error
from glotaran.plugin_system.io_plugin_utils import protect_from_overwrite
from glotaran.utils.ipython import MarkdownStr

if TYPE_CHECKING:
    from os import PathLike
    from typing import Any
    from typing import Callable
    from typing import Literal

    import xarray as xr

    from glotaran.io.interface import DataLoader
    from glotaran.io.interface import DataSaver

DATA_IO_METHODS = ("load_dataset", "save_dataset")
"""Methods used by DataIoInterface plugins."""


def register_data_io(
    format_names: str | list[str],
) -> Callable[[type[DataIoInterface]], type[DataIoInterface]]:
    """Register data io plugins to one or more formats.

    Decorate a data io plugin class with ``@register_data_io(format_name|[*format_names])``
    to add it to the registry.

    Parameters
    ----------
    format_names : str | list[str]
        Name of the data io plugin under which it is registered.

    Returns
    -------
    Callable[[type[DataIoInterface]], type[DataIoInterface]]
        Inner decorator function.

    Examples
    --------
    >>> @register_data_io("my_format_1")
    ... class MyDataIo1(DataIoInterface):
    ...     pass

    >>> @register_data_io(["my_format_1", "my_format_1_alias"])
    ... class MyDataIo2(DataIoInterface):
    ...     pass
    """

    def wrapper(cls: type[DataIoInterface]) -> type[DataIoInterface]:
        add_instantiated_plugin_to_registry(
            plugin_register_keys=format_names,
            plugin_class=cls,
            plugin_registry=__PluginRegistry.data_io,
            plugin_set_func_name="set_data_plugin",
        )
        return cls

    return wrapper


def is_known_data_format(format_name: str) -> bool:
    """Check if a data format is in the data_io registry.

    Parameters
    ----------
    format_name : str
        Name of the data io plugin under which it is registered.

    Returns
    -------
    bool
        Whether or not the data format is a registered data io plugins.
    """
    return is_registered_plugin(
        plugin_register_key=format_name, plugin_registry=__PluginRegistry.data_io
    )


def known_data_formats(full_names: bool = False) -> list[str]:
    """Names of the registered data io plugins.

    Parameters
    ----------
    full_names : bool
        Whether to display the full names the plugins are
        registered under as well.

    Returns
    -------
    list[str]
        List of registered data io plugins.
    """
    return registered_plugins(plugin_registry=__PluginRegistry.data_io, full_names=full_names)


def set_data_plugin(
    format_name: str,
    full_plugin_name: str,
) -> None:
    """Set the plugin used for a specific data format.

    This function is useful when you want to resolve conflicts of installed plugins
    or overwrite the plugin used for a specific format.

    Effected functions:

    - :func:`load_dataset`
    - :func:`save_dataset`

    Parameters
    ----------
    format_name : str
        Format name used to refer to the plugin when used for ``save`` and ``load`` functions.
    full_plugin_name : str
        Full name (import path) of the registered plugin.
    """
    set_plugin(
        plugin_register_key=format_name,
        full_plugin_name=full_plugin_name,
        plugin_registry=__PluginRegistry.data_io,
    )


def get_data_io(format_name: str) -> DataIoInterface:
    """Retrieve a data io plugin from the data_io registry.

    Parameters
    ----------
    format_name : str
        Name of the data io plugin under which it is registered.

    Returns
    -------
    DataIoInterface
        Data io plugin instance.
    """
    return get_plugin_from_registry(
        plugin_register_key=format_name,
        plugin_registry=__PluginRegistry.data_io,
        not_found_error_message=(
            f"Unknown  Data Io format {format_name!r}. Known formats are: {known_data_formats()}"
        ),
    )


@not_implemented_to_value_error
def load_dataset(
    file_name: str | PathLike[str], format_name: str = None, **kwargs: Any
) -> xr.Dataset | xr.DataArray:
    """Read data from a file to :xarraydoc:`Dataset` or :xarraydoc:`DataArray`.

    Parameters
    ----------
    file_name : str | PathLike[str]
        File containing the data.
    format_name : str
        Format the file is in, if not provided it will be inferred from the file extension.
    **kwargs : Any
        Additional keyword arguments passes to the ``read_dataset`` implementation
        of the data io plugin. If you aren't sure about those use ``get_dataloader``
        to get the implementation with the proper help and autocomplete.

    Returns
    -------
    xr.Dataset|xr.DataArray
        Data loaded from the file.
    """
    io = get_data_io(format_name or inferr_file_format(file_name))
    return io.load_dataset(str(file_name), **kwargs)  # type: ignore[call-arg]


@not_implemented_to_value_error
def save_dataset(
    dataset: xr.Dataset | xr.DataArray,
    file_name: str | PathLike[str],
    format_name: str = None,
    *,
    data_filters: list[str] | None = None,
    allow_overwrite: bool = False,
    **kwargs: Any,
) -> None:
    """Save data from :xarraydoc:`Dataset` or :xarraydoc:`DataArray` to a file.

    Parameters
    ----------
    dataset : xr.Dataset | xr.DataArray
        Data to be written to file.
    file_name : str | PathLike[str]
        File to write the data to.
    format_name : str
        Format the file should be in, if not provided it will be inferred from the file extension.
    data_filters : list[str] | None
        Optional list of items in the dataset to be saved.
    allow_overwrite : bool
        Whether or not to allow overwriting existing files, by default False
    **kwargs : Any
        Additional keyword arguments passes to the ``write_dataset`` implementation
        of the data io plugin. If you aren't sure about those use ``get_datawriter``
        to get the implementation with the proper help and autocomplete.
    """
    protect_from_overwrite(file_name, allow_overwrite=allow_overwrite)
    io = get_data_io(format_name or inferr_file_format(file_name, needs_to_exist=False))
    io.save_dataset(  # type: ignore[call-arg]
        file_name=str(file_name),
        dataset=dataset,
        **kwargs,
    )


def get_dataloader(format_name: str) -> DataLoader:
    """Retrieve implementation of the ``read_dataset`` functionality for the format 'format_name'.

    This allows to get the proper help and autocomplete for the function,
    which is especially valuable if the function provides additional options.

    Parameters
    ----------
    format_name : str
        Format the dataloader should be able to read.

    Returns
    -------
    DataLoader
        Function to load data of format ``format_name`` as
        :xarraydoc:`Dataset` or :xarraydoc:`DataArray`.
    """
    io = get_data_io(format_name)
    return get_method_from_plugin(io, "load_dataset")


def get_datasaver(format_name: str) -> DataSaver:
    """Retrieve implementation of the ``save_dataset`` functionality for the format 'format_name'.

    This allows to get the proper help and autocomplete for the function,
    which is especially valuable if the function provides additional options.

    Parameters
    ----------
    format_name : str
        Format the datawriter should be able to write.

    Returns
    -------
    DataSaver
        Function to write :xarraydoc:`Dataset` to the format ``format_name`` .
    """
    io = get_data_io(format_name)
    return get_method_from_plugin(io, "save_dataset")


def show_data_io_method_help(
    format_name: str, method_name: Literal["load_dataset", "save_dataset"]
) -> None:
    """Show help for the implementation of data io plugin methods.

    Parameters
    ----------
    format_name : str
        Format the method should support.
    method_name : {'load_dataset', 'save_dataset'}
        Method name
    """
    io = get_data_io(format_name)
    show_method_help(io, method_name)


def data_io_plugin_table(*, plugin_names: bool = False, full_names: bool = False) -> MarkdownStr:
    """Return registered data io plugins and which functions they support as markdown table.

    This is especially useful when you work with new plugins.

    Parameters
    ----------
    plugin_names : bool
        Whether or not to add the names of the plugins to the table.
    full_names : bool
        Whether to display the full names the plugins are
        registered under as well.

    Returns
    -------
    MarkdownStr
        Markdown table of data io plugins.
    """
    table_data = methods_differ_from_baseclass_table(
        DATA_IO_METHODS,
        known_data_formats(full_names=full_names),
        get_data_io,
        DataIoInterface,
        plugin_names=plugin_names,
    )
    header_values = ["Format name", *DATA_IO_METHODS]
    if plugin_names:
        header_values.append("Plugin name")
    headers = tuple(map(lambda x: f"__{x}__", header_values))
    return MarkdownStr(
        tabulate(
            bool_table_repr(table_data), tablefmt="github", headers=headers, stralign="center"
        )
    )



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from glotaran.io import save_result
from glotaran.project.test.test_result import dummy_result  # noqa: F401

if TYPE_CHECKING:

    from glotaran.project.result import Result


def test_save_result_yml(
    tmp_path: Path,
    dummy_result: Result,  # noqa: F811
):
    """Check all files exist."""

    result_dir = tmp_path / "testresult"
    save_result(result_path=result_dir / "result.yml", result=dummy_result)

    assert (result_dir / "result.md").exists()
    assert (result_dir / "scheme.yml").exists()
    assert (result_dir / "result.yml").exists()
    assert (result_dir / "initial_parameters.csv").exists()
    assert (result_dir / "optimized_parameters.csv").exists()
    assert (result_dir / "dataset1.nc").exists()
    assert (result_dir / "dataset2.nc").exists()
    assert (result_dir / "dataset3.nc").exists()



from __future__ import annotations

import collections
import itertools
from typing import TYPE_CHECKING
from typing import Any
from typing import Deque
from typing import NamedTuple

import numpy as np
import xarray as xr

from glotaran.analysis.optimization_group_calculator import OptimizationGroupCalculator
from glotaran.analysis.util import CalculatedMatrix
from glotaran.analysis.util import apply_weight
from glotaran.analysis.util import calculate_clp_penalties
from glotaran.analysis.util import calculate_matrix
from glotaran.analysis.util import find_closest_index
from glotaran.analysis.util import find_overlap
from glotaran.analysis.util import reduce_matrix
from glotaran.analysis.util import retrieve_clps
from glotaran.model import DatasetModel

if TYPE_CHECKING:
    from glotaran.analysis.optimization_group import OptimizationGroup


class DatasetIndexModel(NamedTuple):
    """A model which contains a dataset label and index information."""

    label: str
    indices: dict[str, int]
    axis: dict[str, np.ndarray]


class DatasetIndexModelGroup(NamedTuple):
    """A model which contains information about a group of dataset with linked clp."""

    data: np.ndarray
    weight: np.ndarray
    has_scaling: bool
    """Indicates if at least one dataset in the group needs scaling."""
    group: str
    """The concatenated labels of the involved datasets."""
    data_sizes: list[int]
    """Holds the sizes of the concatenated datasets."""
    dataset_models: list[DatasetIndexModel]


Bag = Deque[DatasetIndexModelGroup]
"""A deque of dataset group index models."""


class OptimizationGroupCalculatorLinked(OptimizationGroupCalculator):
    """A class to calculate a set of datasets with linked CLP."""

    def __init__(self, group: OptimizationGroup):
        super().__init__(group)
        self._bag = None
        self._full_axis = None

        # TODO: grouping should be user controlled not inferred automatically
        global_dimensions = {d.get_global_dimension() for d in group.dataset_models.values()}
        model_dimensions = {d.get_model_dimension() for d in group.dataset_models.values()}
        if len(global_dimensions) != 1:
            raise ValueError(
                f"Cannot group datasets. Global dimensions '{global_dimensions}' do not match."
            )
        if len(model_dimensions) != 1:
            raise ValueError(
                f"Cannot group datasets. Model dimension '{model_dimensions}' do not match."
            )
        self._index_dependent = any(d.is_index_dependent() for d in group.dataset_models.values())
        self._global_dimension = global_dimensions.pop()
        self._model_dimension = model_dimensions.pop()
        self._group_clp_labels = None
        self._groups = None
        self._has_weights = any("weight" in d for d in group.data.values())

    @property
    def bag(self) -> Bag:
        if self._bag is None:
            self.init_bag()
        return self._bag

    def init_bag(self):
        """Initializes a grouped problem bag."""
        self._bag = None
        datasets = None
        for label, dataset_model in self._group.dataset_models.items():

            data = dataset_model.get_data()
            weight = dataset_model.get_weight()
            if weight is None and self._has_weights:
                weight = np.ones_like(data)

            global_axis = dataset_model.get_global_axis()
            model_axis = dataset_model.get_model_axis()
            has_scaling = dataset_model.scale is not None

            if self._bag is None:
                self._bag = collections.deque(
                    DatasetIndexModelGroup(
                        data=data[:, i],
                        weight=weight[:, i] if weight is not None else None,
                        has_scaling=has_scaling,
                        group=label,
                        data_sizes=[model_axis.size],
                        dataset_models=[
                            DatasetIndexModel(
                                label,
                                {
                                    self._global_dimension: i,
                                },
                                {
                                    self._model_dimension: model_axis,
                                    self._global_dimension: global_axis,
                                },
                            )
                        ],
                    )
                    for i, value in enumerate(global_axis)
                )
                datasets = collections.deque([label] for _, _ in enumerate(global_axis))
                self._full_axis = collections.deque(global_axis)
            else:
                self._append_to_grouped_bag(
                    label, datasets, global_axis, model_axis, data, weight, has_scaling
                )
        self._full_axis = np.asarray(self._full_axis)
        self._groups = {"".join(d): d for d in datasets}

    def _append_to_grouped_bag(
        self,
        label: str,
        datasets: Deque[str],
        global_axis: np.ndarray,
        model_axis: np.ndarray,
        data: xr.DataArray,
        weight: xr.DataArray,
        has_scaling: bool,
    ):
        i1, i2 = find_overlap(self._full_axis, global_axis, atol=self._group._clp_link_tolerance)

        for i, j in enumerate(i1):
            datasets[j].append(label)
            data_stripe = data[:, i2[i]]
            self._bag[j] = DatasetIndexModelGroup(
                data=np.concatenate(
                    [
                        self._bag[j].data,
                        data_stripe,
                    ]
                ),
                weight=np.concatenate([self._bag[j].weight, weight[:, i2[i]]])
                if weight is not None
                else None,
                has_scaling=has_scaling or self._bag[j].has_scaling,
                group=self._bag[j].group + label,
                data_sizes=self._bag[j].data_sizes + [data_stripe.size],
                dataset_models=self._bag[j].dataset_models
                + [
                    DatasetIndexModel(
                        label,
                        {
                            self._global_dimension: i2[i],
                        },
                        {
                            self._model_dimension: model_axis,
                            self._global_dimension: global_axis,
                        },
                    )
                ],
            )

        # Add non-overlaping regions
        begin_overlap = i2[0] if len(i2) != 0 else 0
        end_overlap = i2[-1] + 1 if len(i2) != 0 else 0
        for i in itertools.chain(range(begin_overlap), range(end_overlap, len(global_axis))):
            data_stripe = data[:, i]
            problem = DatasetIndexModelGroup(
                data=data_stripe,
                weight=weight[:, i] if weight is not None else None,
                has_scaling=has_scaling,
                group=label,
                data_sizes=[data_stripe.size],
                dataset_models=[
                    DatasetIndexModel(
                        label,
                        {
                            self._global_dimension: i,
                        },
                        {
                            self._model_dimension: model_axis,
                            self._global_dimension: global_axis,
                        },
                    )
                ],
            )
            if i < end_overlap:
                datasets.appendleft([label])
                self._full_axis.appendleft(global_axis[i])
                self._bag.appendleft(problem)
            else:
                datasets.append([label])
                self._full_axis.append(global_axis[i])
                self._bag.append(problem)

    @property
    def groups(self) -> dict[str, list[str]]:
        if not self._groups:
            self.init_bag()
        return self._groups

    def calculate_matrices(self):
        if self._index_dependent:
            self.calculate_index_dependent_matrices()
        else:
            self.calculate_index_independent_matrices()

    def calculate_index_dependent_matrices(
        self,
    ) -> tuple[dict[str, list[CalculatedMatrix]], list[CalculatedMatrix],]:
        """Calculates the index dependent model matrices."""

        def calculate_group(
            group_model: DatasetIndexModelGroup, dataset_models: dict[str, DatasetModel]
        ) -> tuple[list[CalculatedMatrix], list[str], CalculatedMatrix]:
            matrices = [
                calculate_matrix(
                    dataset_models[dataset_index_model.label],
                    dataset_index_model.indices,
                )
                for dataset_index_model in group_model.dataset_models
            ]
            global_index = group_model.dataset_models[0].indices[self._global_dimension]
            global_index = group_model.dataset_models[0].axis[self._global_dimension][global_index]
            combined_matrix = combine_matrices(matrices)
            group_clp_labels = combined_matrix.clp_labels
            reduced_matrix = reduce_matrix(
                combined_matrix, self._group.model, self._group.parameters, global_index
            )
            return matrices, group_clp_labels, reduced_matrix

        results = list(
            map(
                lambda group_model: calculate_group(group_model, self._group.dataset_models),
                self.bag,
            )
        )

        matrices = list(map(lambda result: result[0], results))

        self._group._matrices = {}
        for i, group_model in enumerate(self._bag):
            for j, dataset_index_model in enumerate(group_model.dataset_models):
                if dataset_index_model.label not in self._group._matrices:
                    self._group._matrices[dataset_index_model.label] = []
                self._group._matrices[dataset_index_model.label].append(matrices[i][j])

        self._group_clp_labels = list(map(lambda result: result[1], results))
        self._group._reduced_matrices = list(map(lambda result: result[2], results))
        return self._group._matrices, self._group._reduced_matrices

    def calculate_index_independent_matrices(
        self,
    ) -> tuple[dict[str, CalculatedMatrix], dict[str, CalculatedMatrix],]:
        """Calculates the index independent model matrices."""
        self._group._matrices = {}
        self._group._reduced_matrices = {}
        self._group_clp_labels = {}

        for label, dataset_model in self._group.dataset_models.items():
            self._group._matrices[label] = calculate_matrix(
                dataset_model,
                {},
            )
            self._group_clp_labels[label] = self._group._matrices[label].clp_labels
            self._group._reduced_matrices[label] = reduce_matrix(
                self._group._matrices[label],
                self._group.model,
                self._group.parameters,
                None,
            )

        for group_label, group in self.groups.items():
            if group_label not in self._group._matrices:
                self._group._reduced_matrices[group_label] = combine_matrices(
                    [self._group._reduced_matrices[label] for label in group]
                )
                group_clp_labels = []
                for label in group:
                    for clp_label in self._group._matrices[label].clp_labels:
                        if clp_label not in group_clp_labels:
                            group_clp_labels.append(clp_label)
                self._group_clp_labels[group_label] = group_clp_labels

        return self._group._matrices, self._group._reduced_matrices

    def calculate_residual(self):
        results = (
            list(
                map(
                    self._index_dependent_residual,
                    self.bag,
                    self._group.reduced_matrices,
                    self._group_clp_labels,
                    self._full_axis,
                )
            )
            if self._index_dependent
            else list(map(self._index_independent_residual, self.bag, self._full_axis))
        )

        self._clp_labels = list(map(lambda result: result[0], results))
        self._grouped_clps = list(map(lambda result: result[1], results))

        self._group._weighted_residuals = list(map(lambda result: result[2], results))
        self._group._residuals = list(map(lambda result: result[3], results))
        self._group._additional_penalty = calculate_clp_penalties(
            self._group.model,
            self._group.parameters,
            self._clp_labels,
            self._grouped_clps,
            self._full_axis,
            self._group.dataset_models,
        )

        return (
            self._group._reduced_clps,
            self._group._clps,
            self._group._weighted_residuals,
            self._group._residuals,
        )

    def _index_dependent_residual(
        self,
        group_model: DatasetIndexModelGroup,
        matrix: CalculatedMatrix,
        clp_labels: str,
        index: Any,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:

        reduced_clp_labels = matrix.clp_labels
        matrix = matrix.matrix
        if group_model.weight is not None:
            apply_weight(matrix, group_model.weight)
        data = group_model.data

        self._apply_scale(group_model, matrix)

        reduced_clps, weighted_residual = self._group._residual_function(matrix, data)
        clps = retrieve_clps(
            self._group.model,
            self._group.parameters,
            clp_labels,
            reduced_clp_labels,
            reduced_clps,
            index,
        )
        residual = (
            weighted_residual / group_model.weight
            if group_model.weight is not None
            else weighted_residual
        )
        return clp_labels, clps, weighted_residual, residual

    def _index_independent_residual(self, group_model: DatasetIndexModelGroup, index: Any):
        matrix = self._group.reduced_matrices[group_model.group]
        reduced_clp_labels = matrix.clp_labels
        matrix = matrix.matrix.copy()
        if group_model.weight is not None:
            apply_weight(matrix, group_model.weight)
        data = group_model.data

        self._apply_scale(group_model, matrix)

        reduced_clps, weighted_residual = self._group._residual_function(matrix, data)
        clp_labels = self._group_clp_labels[group_model.group]
        clps = retrieve_clps(
            self._group.model,
            self._group.parameters,
            clp_labels,
            reduced_clp_labels,
            reduced_clps,
            index,
        )
        residual = (
            weighted_residual / group_model.weight
            if group_model.weight is not None
            else weighted_residual
        )
        return clp_labels, clps, weighted_residual, residual

    def _apply_scale(self, group_model: DatasetIndexModelGroup, matrix: np.ndarray):
        if group_model.has_scaling:
            for i, index_model in enumerate(group_model.dataset_models):
                label = index_model.label
                if self._group.dataset_models[label] is not None:
                    start = sum(group_model.data_sizes[0:i])
                    end = start + group_model.data_sizes[i]
                    matrix[start:end, :] *= self._group.dataset_models[label].scale

    def prepare_result_creation(self):
        if self._group._residuals is None:
            self.calculate_residual()
        full_clp_labels = self._clp_labels
        full_clps = self._grouped_clps
        self._group._clps = {}
        for label, matrix in self._group.matrices.items():
            # TODO deal with different clps at indices
            clp_labels = matrix[0].clp_labels if self._index_dependent else matrix.clp_labels

            # find offset in the full axis
            global_axis = self._group.dataset_models[label].get_global_axis()
            offset = find_closest_index(global_axis[0], self._full_axis)

            clps = []
            for i in range(global_axis.size):
                full_index_clp_labels = full_clp_labels[i + offset]
                index_clps = full_clps[i + offset]
                mask = [full_index_clp_labels.index(clp_label) for clp_label in clp_labels]
                clps.append(index_clps[mask])

            self._group._clps[label] = xr.DataArray(
                clps,
                coords=((self._global_dimension, global_axis), ("clp_label", clp_labels)),
            )

    def create_index_dependent_result_dataset(self, label: str, dataset: xr.Dataset) -> xr.Dataset:
        """Creates a result datasets for index dependent matrices."""

        for index, grouped_problem in enumerate(self.bag):

            if label in grouped_problem.group:
                group_index = [
                    descriptor.label for descriptor in grouped_problem.dataset_models
                ].index(label)
                group_descriptor = grouped_problem.dataset_models[group_index]
                global_index = group_descriptor.indices[self._global_dimension]
                global_index = group_descriptor.axis[self._global_dimension][global_index]

                self._add_grouped_residual_to_dataset(
                    dataset, grouped_problem, index, group_index, global_index
                )

        dataset["matrix"] = (
            (
                (self._global_dimension),
                (self._model_dimension),
                ("clp_label"),
            ),
            np.asarray([m.matrix for m in self._group.matrices[label]]),
        )
        dataset["clp"] = self._group.clps[label]

        return dataset

    def create_index_independent_result_dataset(
        self, label: str, dataset: xr.Dataset
    ) -> xr.Dataset:
        """Creates a result datasets for index independent matrices."""

        dataset["matrix"] = (
            (
                (self._model_dimension),
                ("clp_label"),
            ),
            self._group.matrices[label].matrix,
        )
        dataset["clp"] = self._group.clps[label]

        for index, grouped_problem in enumerate(self.bag):

            if label in grouped_problem.group:
                group_index = [
                    descriptor.label for descriptor in grouped_problem.dataset_models
                ].index(label)
                group_descriptor = grouped_problem.dataset_models[group_index]
                global_index = group_descriptor.indices[self._global_dimension]
                global_index = group_descriptor.axis[self._global_dimension][global_index]

                self._add_grouped_residual_to_dataset(
                    dataset, grouped_problem, index, group_index, global_index
                )

        return dataset

    def _add_grouped_residual_to_dataset(
        self,
        dataset: xr.Dataset,
        grouped_problem: DatasetIndexModelGroup,
        index: int,
        group_index: int,
        global_index: int,
    ):
        if "residual" not in dataset:
            dim1 = dataset.coords[self._model_dimension].size
            dim2 = dataset.coords[self._global_dimension].size
            dataset["weighted_residual"] = (
                (self._model_dimension, self._global_dimension),
                np.zeros((dim1, dim2), dtype=np.float64),
            )
            dataset["residual"] = (
                (self._model_dimension, self._global_dimension),
                np.zeros((dim1, dim2), dtype=np.float64),
            )

        start = sum(
            self._group.data[grouped_problem.dataset_models[i].label]
            .coords[self._model_dimension]
            .size
            for i in range(group_index)
        )

        end = start + dataset.coords[self._model_dimension].size
        dataset.weighted_residual.loc[
            {self._global_dimension: global_index}
        ] = self._group.weighted_residuals[index][start:end]
        dataset.residual.loc[{self._global_dimension: global_index}] = self._group.residuals[
            index
        ][start:end]

    def calculate_full_penalty(self) -> np.ndarray:
        if self._group._full_penalty is None:
            residuals = self._group.weighted_residuals
            additional_penalty = self._group.additional_penalty

            self._group._full_penalty = (
                np.concatenate((np.concatenate(residuals), additional_penalty))
                if additional_penalty is not None
                else np.concatenate(residuals)
            )
        return self._group._full_penalty


def combine_matrices(matrices: list[CalculatedMatrix]) -> CalculatedMatrix:
    masks = []
    full_clp_labels = None
    sizes = []
    dim1 = 0
    for matrix in matrices:
        clp_labels = matrix.clp_labels
        model_axis_size = matrix.matrix.shape[0]
        sizes.append(model_axis_size)
        dim1 += model_axis_size
        if full_clp_labels is None:
            full_clp_labels = clp_labels.copy()
            masks.append([i for i, _ in enumerate(clp_labels)])
        else:
            mask = []
            for c in clp_labels:
                if c not in full_clp_labels:
                    full_clp_labels.append(c)
                mask.append(full_clp_labels.index(c))
            masks.append(mask)
    dim2 = len(full_clp_labels)
    full_matrix = np.zeros((dim1, dim2), dtype=np.float64)
    start = 0
    for i, m in enumerate(matrices):
        end = start + sizes[i]
        full_matrix[start:end, masks[i]] = m.matrix
        start = end

    return CalculatedMatrix(full_clp_labels, full_matrix)



from copy import deepcopy

import pytest

from glotaran.analysis.optimization_group import OptimizationGroup
from glotaran.analysis.simulation import simulate
from glotaran.analysis.test.models import TwoCompartmentDecay as suite
from glotaran.model import Relation
from glotaran.parameter import ParameterGroup
from glotaran.project import Scheme


@pytest.mark.parametrize("index_dependent", [True, False])
@pytest.mark.parametrize("link_clp", [True, False])
def test_relations(index_dependent, link_clp):
    model = deepcopy(suite.model)
    model.dataset_group_models["default"].link_clp = link_clp
    model.megacomplex["m1"].is_index_dependent = index_dependent
    model.clp_relations.append(
        Relation.from_dict({"source": "s1", "target": "s2", "parameter": "3"})
    )
    parameters = ParameterGroup.from_list([11e-4, 22e-5, 2])

    print("link_clp", link_clp, "index_dependent", index_dependent)
    dataset = simulate(
        suite.sim_model,
        "dataset1",
        parameters,
        {"global": suite.global_axis, "model": suite.model_axis},
    )
    scheme = Scheme(model=model, parameters=parameters, data={"dataset1": dataset})
    optimization_group = OptimizationGroup(scheme, model.get_dataset_groups()["default"])

    if index_dependent:
        reduced_matrix = (
            optimization_group.reduced_matrices[0]
            if link_clp
            else optimization_group.reduced_matrices["dataset1"][0]
        )
    else:
        reduced_matrix = optimization_group.reduced_matrices["dataset1"]
    matrix = (
        optimization_group.matrices["dataset1"][0]
        if index_dependent
        else optimization_group.matrices["dataset1"]
    )

    result_data = optimization_group.create_result_data()
    print(result_data)
    clps = result_data["dataset1"].clp

    assert "s2" not in reduced_matrix.clp_labels
    assert "s2" in clps.coords["clp_label"]
    assert clps.sel(clp_label="s2") == clps.sel(clp_label="s1") * 2
    assert "s2" in matrix.clp_labels



""" Glotaran Relation """
from __future__ import annotations

from glotaran.model.interval_property import IntervalProperty
from glotaran.model.item import model_item
from glotaran.parameter import Parameter


@model_item(
    properties={
        "source": str,
        "target": str,
        "parameter": Parameter,
    },
    has_label=False,
)
class Relation(IntervalProperty):
    """Applies a relation between clps as

    :math:`target = parameter * source`.
    """



"""Test deprecated imports from 'glotaran/__init__.py' """
from __future__ import annotations

from typing import TYPE_CHECKING
from warnings import warn

import pytest

from glotaran import read_model_from_yaml
from glotaran import read_model_from_yaml_file
from glotaran import read_parameters_from_csv_file
from glotaran import read_parameters_from_yaml
from glotaran import read_parameters_from_yaml_file
from glotaran.deprecation.deprecation_utils import GlotaranApiDeprecationWarning
from glotaran.deprecation.modules.test import deprecation_warning_on_call_test_helper
from glotaran.model import Model
from glotaran.parameter import ParameterGroup

if TYPE_CHECKING:
    from pathlib import Path


def dummy_warn(foo, bar=False):
    warn(GlotaranApiDeprecationWarning("foo"), stacklevel=2)
    if not isinstance(bar, bool):
        raise ValueError("not a bool")
    return foo, bar


def dummy_no_warn(foo, bar=False):
    return foo, bar


def test_deprecation_warning_on_call_test_helper():
    """Correct result passed on"""
    record, result = deprecation_warning_on_call_test_helper(
        dummy_warn, args=["foo"], kwargs={"bar": True}
    )
    assert len(record) == 1
    assert result == ("foo", True)


def test_deprecation_warning_on_call_test_helper_error_reraise():
    """Raise if raise_exception and args or kwargs"""

    with pytest.raises(ValueError, match="not a bool"):
        deprecation_warning_on_call_test_helper(
            dummy_warn, args=["foo"], kwargs={"bar": "baz"}, raise_exception=True
        )


@pytest.mark.xfail(strict=True, reason="Function did not warn.")
def test_deprecation_warning_on_call_test_helper_no_warn():
    """Fail no warning"""
    deprecation_warning_on_call_test_helper(dummy_no_warn, args=["foo"], kwargs={"bar": True})


def test_read_model_from_yaml():
    """read_model_from_yaml raises warning"""
    yaml = """
    type: kinetic-spectrum
    megacomplex: {}
    """
    _, result = deprecation_warning_on_call_test_helper(
        read_model_from_yaml, args=[yaml], raise_exception=True
    )

    assert isinstance(result, Model)


def test_read_model_from_yaml_file(tmp_path: Path):
    """read_model_from_yaml_file raises warning"""
    yaml = """
    type: kinetic-spectrum
    megacomplex: {}
    """
    model_file = tmp_path / "model.yaml"
    model_file.write_text(yaml)
    _, result = deprecation_warning_on_call_test_helper(
        read_model_from_yaml_file, args=[str(model_file)], raise_exception=True
    )

    assert isinstance(result, Model)


def test_read_parameters_from_csv_file(tmp_path: Path):
    """read_parameters_from_csv_file raises warning"""
    parameters_file = tmp_path / "parameters.csv"
    parameters_file.write_text("label,value\nfoo,123")
    _, result = deprecation_warning_on_call_test_helper(
        read_parameters_from_csv_file,
        args=[str(parameters_file)],
        raise_exception=True,
    )

    assert isinstance(result, ParameterGroup)
    assert result.get("foo") == 123


def test_read_parameters_from_yaml():
    """read_parameters_from_yaml raises warning"""
    _, result = deprecation_warning_on_call_test_helper(
        read_parameters_from_yaml, args=["foo:\n  - 123"], raise_exception=True
    )

    assert isinstance(result, ParameterGroup)
    assert isinstance(result["foo"], ParameterGroup)
    assert result["foo"].get("1") == 123


def test_read_parameters_from_yaml_file(tmp_path: Path):
    """read_parameters_from_yaml_file raises warning"""
    parameters_file = tmp_path / "parameters.yaml"
    parameters_file.write_text("foo:\n  - 123")
    _, result = deprecation_warning_on_call_test_helper(
        read_parameters_from_yaml_file, args=[str(parameters_file)], raise_exception=True
    )

    assert isinstance(result, ParameterGroup)
    assert isinstance(result["foo"], ParameterGroup)
    assert result["foo"].get("1") == 123



"""Deprecation functions for the yaml parser."""
from __future__ import annotations

from typing import TYPE_CHECKING

from glotaran.deprecation import deprecate_dict_entry

if TYPE_CHECKING:
    from typing import Any
    from typing import MutableMapping


def model_spec_deprecations(spec: MutableMapping[Any, Any]) -> None:
    """Check deprecations in the model specification ``spec`` dict.

    Parameters
    ----------
    spec : MutableMapping[Any, Any]
        Model specification dictionary
    """
    load_model_stack_level = 7
    deprecate_dict_entry(
        dict_to_check=spec,
        deprecated_usage="type: kinetic-spectrum",
        new_usage="default_megacomplex: decay",
        to_be_removed_in_version="0.7.0",
        replace_rules=({"type": "kinetic-spectrum"}, {"default_megacomplex": "decay"}),
        stacklevel=load_model_stack_level,
    )

    deprecate_dict_entry(
        dict_to_check=spec,
        deprecated_usage="type: spectral-model",
        new_usage="default_megacomplex: spectral",
        to_be_removed_in_version="0.7.0",
        replace_rules=({"type": "spectral-model"}, {"default_megacomplex": "spectral"}),
        stacklevel=load_model_stack_level,
    )

    deprecate_dict_entry(
        dict_to_check=spec,
        deprecated_usage="default-megacomplex",
        new_usage="default_megacomplex",
        to_be_removed_in_version="0.7.0",
        swap_keys=("default-megacomplex", "default_megacomplex"),
        stacklevel=load_model_stack_level,
    )

    deprecate_dict_entry(
        dict_to_check=spec,
        deprecated_usage="spectral_relations",
        new_usage="clp_relations",
        to_be_removed_in_version="0.7.0",
        swap_keys=("spectral_relations", "clp_relations"),
        stacklevel=load_model_stack_level,
    )

    deprecate_dict_entry(
        dict_to_check=spec,
        deprecated_usage="relations",
        new_usage="clp_relations",
        to_be_removed_in_version="0.7.0",
        swap_keys=("relations", "clp_relations"),
        stacklevel=load_model_stack_level,
    )

    if "clp_relations" in spec:
        for relation in spec["clp_relations"]:
            deprecate_dict_entry(
                dict_to_check=relation,
                deprecated_usage="clp_relations:\n - compartment",
                new_usage="clp_relations:\n - source",
                to_be_removed_in_version="0.7.0",
                swap_keys=("compartment", "source"),
                stacklevel=load_model_stack_level,
            )

    deprecate_dict_entry(
        dict_to_check=spec,
        deprecated_usage="spectral_constraints",
        new_usage="clp_constraints",
        to_be_removed_in_version="0.7.0",
        swap_keys=("spectral_constraints", "clp_constraints"),
        stacklevel=load_model_stack_level,
    )

    deprecate_dict_entry(
        dict_to_check=spec,
        deprecated_usage="constraints",
        new_usage="clp_constraints",
        to_be_removed_in_version="0.7.0",
        swap_keys=("constraints", "clp_constraints"),
        stacklevel=load_model_stack_level,
    )

    if "clp_constraints" in spec:
        for constraint in spec["clp_constraints"]:
            deprecate_dict_entry(
                dict_to_check=constraint,
                deprecated_usage="clp_constraints:\n - compartment",
                new_usage="clp_constraints:\n - target",
                to_be_removed_in_version="0.7.0",
                swap_keys=("compartment", "target"),
                stacklevel=load_model_stack_level,
            )

    deprecate_dict_entry(
        dict_to_check=spec,
        deprecated_usage="equal_area_penalties",
        new_usage="clp_area_penalties",
        to_be_removed_in_version="0.7.0",
        swap_keys=("equal_area_penalties", "clp_area_penalties"),
        stacklevel=load_model_stack_level,
    )

    if "irf" in spec:
        for _, irf in spec["irf"].items():
            deprecate_dict_entry(
                dict_to_check=irf,
                deprecated_usage="center_dispersion",
                new_usage="center_dispersion_coefficients",
                to_be_removed_in_version="0.7.0",
                swap_keys=("center_dispersion", "center_dispersion_coefficients"),
                stacklevel=load_model_stack_level,
            )

        for _, irf in spec["irf"].items():
            deprecate_dict_entry(
                dict_to_check=irf,
                deprecated_usage="width_dispersion",
                new_usage="width_dispersion_coefficients",
                to_be_removed_in_version="0.7.0",
                swap_keys=("width_dispersion", "width_dispersion_coefficients"),
                stacklevel=load_model_stack_level,
            )


def scheme_spec_deprecations(spec: MutableMapping[Any, Any]) -> None:
    """Check deprecations in the scheme specification ``spec`` dict.

    Parameters
    ----------
    spec : MutableMapping[Any, Any]
        Scheme specification dictionary
    """
    load_scheme_stack_level = 7

    deprecate_dict_entry(
        dict_to_check=spec,
        deprecated_usage="maximum-number-function-evaluations: <number>",
        new_usage="maximum_number_function_evaluations: <number>",
        to_be_removed_in_version="0.7.0",
        swap_keys=("maximum-number-function-evaluations", "maximum_number_function_evaluations"),
        stacklevel=load_scheme_stack_level,
    )

    deprecate_dict_entry(
        dict_to_check=spec,
        deprecated_usage="non-negative-least-squares",
        new_usage=("<model_file>dataset_groups.default.residual_function"),
        to_be_removed_in_version="0.7.0",
        swap_keys=("non-negative-least-squares", "non_negative_least_squares"),
        stacklevel=load_scheme_stack_level,
    )



from __future__ import annotations

from copy import deepcopy

import pytest
from rich import pretty
from rich import print  # pylint: disable=W0622

from glotaran.model import Model
from glotaran.parameter import ParameterGroup
from glotaran.testing.model_generators import SimpleModelGenerator

pretty.install()


REF_PARAMETER_DICT = {
    "rates": [
        ["1", 501e-3],
        ["2", 202e-4],
        ["3", 105e-5],
        {"non-negative": True},
    ],
    "irf": [["center", 1.3], ["width", 7.8]],
    "inputs": [
        ["1", 1],
        ["0", 0],
        {"vary": False},
    ],
}

REF_MODEL_DICT = {
    "default_megacomplex": "decay",
    "initial_concentration": {
        "j1": {
            "compartments": ["s1", "s2", "s3"],
            "parameters": ["inputs.1", "inputs.0", "inputs.0"],
        },
    },
    "megacomplex": {
        "mc1": {"k_matrix": ["k1"]},
    },
    "k_matrix": {
        "k1": {
            "matrix": {
                ("s2", "s1"): "rates.1",
                ("s3", "s2"): "rates.2",
                ("s3", "s3"): "rates.3",
            }
        }
    },
    "irf": {
        "irf1": {
            "type": "multi-gaussian",
            "center": ["irf.center"],
            "width": ["irf.width"],
        },
    },
    "dataset": {
        "dataset1": {
            "initial_concentration": "j1",
            "irf": "irf1",
            "megacomplex": ["mc1"],
        },
    },
}


def simple_diff_between_string(string1, string2):
    return "".join(c2 for c1, c2 in zip(string1, string2) if c1 != c2)


def test_three_component_sequential_model():
    ref_model = Model.from_dict(deepcopy(REF_MODEL_DICT))
    ref_parameters = ParameterGroup.from_dict(deepcopy(REF_PARAMETER_DICT))
    generator = SimpleModelGenerator(
        rates=[501e-3, 202e-4, 105e-5, {"non-negative": True}],
        irf={"center": 1.3, "width": 7.8},
        k_matrix="sequential",
    )
    for key, _ in REF_PARAMETER_DICT.items():
        assert key in generator.parameters_dict
        # TODO: check contents

    model, parameters = generator.model_and_parameters
    assert str(ref_model) == str(model), print(
        simple_diff_between_string(str(model), str(ref_model))
    )
    assert str(ref_parameters) == str(parameters), print(
        simple_diff_between_string(str(parameters), str(ref_parameters))
    )


def test_only_rates_no_irf():
    generator = SimpleModelGenerator(rates=[0.1, 0.02, 0.003])
    assert "irf" not in generator.model_dict.keys()


def test_no_rates():
    generator = SimpleModelGenerator()
    assert generator.valid is False


def test_one_rate():
    generator = SimpleModelGenerator([1])
    assert generator.valid is True
    assert "is valid" in generator.validate()


def test_rates_not_a_list():
    generator = SimpleModelGenerator(1)
    assert generator.valid is False
    with pytest.raises(ValueError):
        print(generator.validate())


def test_set_rates_delayed():
    generator = SimpleModelGenerator()
    generator.rates = [1, 2, 3]
    assert generator.valid is True



"""Glotaran Model Package

This package contains the Glotaran's base model object, the model decorators and
common model items.
"""

from glotaran.model.clp_penalties import EqualAreaPenalty
from glotaran.model.constraint import Constraint
from glotaran.model.constraint import OnlyConstraint
from glotaran.model.constraint import ZeroConstraint
from glotaran.model.dataset_group import DatasetGroup
from glotaran.model.dataset_group import DatasetGroupModel
from glotaran.model.dataset_model import DatasetModel
from glotaran.model.item import model_item
from glotaran.model.item import model_item_typed
from glotaran.model.megacomplex import Megacomplex
from glotaran.model.megacomplex import megacomplex
from glotaran.model.model import Model
from glotaran.model.relation import Relation
from glotaran.model.util import ModelError
from glotaran.model.weight import Weight
from glotaran.plugin_system.megacomplex_registration import get_megacomplex
from glotaran.plugin_system.megacomplex_registration import is_known_megacomplex
from glotaran.plugin_system.megacomplex_registration import known_megacomplex_names
from glotaran.plugin_system.megacomplex_registration import megacomplex_plugin_table
from glotaran.plugin_system.megacomplex_registration import set_megacomplex_plugin



from __future__ import annotations

import pickle

import numpy as np
import pytest

from glotaran.io import load_parameters
from glotaran.io import save_parameters
from glotaran.parameter import Parameter


@pytest.mark.parametrize("label, expected", (("foo", "foo"), (0, "0"), (1, "1"), (None, None)))
def test_parameter_label_always_str_or_None(label: str | int | None, expected: str | None):
    """Parameter.label is always a string or None"""
    parameter = Parameter(label=label)  # type:ignore[arg-type]
    assert parameter.label == expected


@pytest.mark.parametrize(
    "label",
    (2.0, np.nan, "foo.bar"),
)
def test_parameter_label_error_wrong_label_pattern(label: str | int | float):
    """Error if label can't be casted to a valid label str"""
    with pytest.raises(ValueError, match=f"'{label}' is not a valid group label."):
        Parameter(label=label)  # type:ignore[arg-type]


def test_param_repr():
    """Repr creates code to recreate the object."""
    result = Parameter(label="foo", value=1.0, expression="$foo.bar", vary=False)
    result_short = Parameter(label="foo", value=1, expression="$foo.bar")
    expected = "Parameter(label='foo', value=1.0, expression='$foo.bar', vary=False)"

    assert result == result_short
    assert result.vary == result_short.vary
    assert result.__repr__() == expected
    assert result_short.__repr__() == expected


def test_param_array():
    params = """
    - 5
    - 4
    - 3
    - 2
    - 1
    """

    params = load_parameters(params, format_name="yml_str")

    assert len(list(params.all())) == 5

    assert [p.label for _, p in params.all()] == [f"{i}" for i in range(1, 6)]
    assert [p.value for _, p in params.all()] == list(range(1, 6))[::-1]


def test_param_scientific():
    values = [5e3, -4.2e-4, 3e-2, -2e6]
    params = """
    - ["1", 5e3]
    - ["2", -4.2e-4]
    - ["3", 3e-2]
    - ["4", -2e6]
    """

    params = load_parameters(params, format_name="yml_str")

    assert [p.value for _, p in params.all()] == values


def test_param_label():
    params = """
    - ["5", 1]
    - ["4", 2]
    - ["3", 3]
    """

    params = load_parameters(params, format_name="yml_str")

    assert len(list(params.all())) == 3
    assert [p.label for _, p in params.all()] == [f"{i}" for i in range(5, 2, -1)]
    assert [p.value for _, p in params.all()] == list(range(1, 4))


def test_param_group_copy():
    params = """
    kinetic:
        - ["5", 1, {non-negative: true, min: -1, max: 1, vary: false}]
        - 4
        - 5
    j:
        - 7
        - 8
    """
    params = load_parameters(params, format_name="yml_str")
    copy = params.copy()

    for label, parameter in params.all():
        assert copy.has(label)
        copied_parameter = copy.get(label)
        assert parameter.value == copied_parameter.value
        assert parameter.non_negative == copied_parameter.non_negative
        assert parameter.minimum == copied_parameter.minimum
        assert parameter.maximum == copied_parameter.maximum
        assert parameter.vary == copied_parameter.vary


def test_param_options():
    params = """
    - ["5", 1, {non-negative: false, min: -1, max: 1, vary: false}]
    - ["6", 4e2, {non-negative: true, min: -7e2, max: 8e2, vary: true}]
    - ["7", 2e4]
    """

    params = load_parameters(params, format_name="yml_str")

    assert params.get("5").value == 1.0
    assert not params.get("5").non_negative
    assert params.get("5").minimum == -1
    assert params.get("5").maximum == 1
    assert not params.get("5").vary

    assert params.get("6").value == 4e2
    assert params.get("6").non_negative
    assert params.get("6").minimum == -7e2
    assert params.get("6").maximum == 8e2
    assert params.get("6").vary

    assert params.get("7").value == 2e4
    assert not params.get("7").non_negative
    assert params.get("7").minimum == float("-inf")
    assert params.get("7").maximum == float("inf")
    assert params.get("7").vary


def test_param_block_options():
    params = """
    block:
        - 1.0
        - [3.4, {vary: true}]
        - {vary: false}
    """

    params = load_parameters(params, format_name="yml_str")
    assert not params.get("block.1").vary
    assert params.get("block.2").vary


def test_nested_param_list():
    params = """
    kinetic:
        - 3
        - 4
        - 5
    j:
        - 7
        - 8
    """

    params = load_parameters(params, format_name="yml_str")

    assert len(list(params.all())) == 5
    group = params["kinetic"]
    assert len(list(group.all())) == 3
    assert [p.label for _, p in group.all()] == [f"{i}" for i in range(1, 4)]
    assert [p.value for _, p in group.all()] == list(range(3, 6))
    group = params["j"]
    assert len(list(group.all())) == 2
    assert [p.label for _, p in group.all()] == [f"{i}" for i in range(1, 3)]
    assert [p.value for _, p in group.all()] == list(range(7, 9))


def test_nested_param_group():
    params = """
    kinetic:
        j:
            - 7
            - 8
            - 9
    """

    params = load_parameters(params, format_name="yml_str")
    assert len(list(params.all())) == 3
    group = params["kinetic"]
    assert len(list(group.all())) == 3
    group = group["j"]
    assert len(list(group.all())) == 3
    assert [p.label for _, p in group.all()] == [f"{i}" for i in range(1, 4)]
    assert [p.value for _, p in group.all()] == list(range(7, 10))


def test_parameter_set_from_group():
    """Parameter extracted from group has correct values"""
    group = load_parameters(
        "foo:\n  - [\"1\", 123,{non-negative: true, min: 10, max: 8e2, vary: true, expr:'2'}]",
        format_name="yml_str",
    )
    parameter = Parameter(full_label="foo.1")
    parameter.set_from_group(group=group)

    assert parameter.value == 123
    assert parameter.non_negative is True
    assert np.allclose(parameter.minimum, 10)
    assert np.allclose(parameter.maximum, 800)
    assert parameter.vary is True
    # Set to None since value and expr were provided?
    assert parameter.expression is None


def test_parameter_value_not_numeric_error():
    """Error if value isn't numeric."""
    with pytest.raises(TypeError, match="Parameter value must be numeric"):
        Parameter(value="foo")  # type:ignore[arg-type]


def test_parameter_maximum_not_numeric_error():
    """Error if maximum isn't numeric."""
    with pytest.raises(TypeError, match="Parameter maximum must be numeric"):
        Parameter(maximum="foo")  # type:ignore[arg-type]


def test_parameter_minimum_not_numeric_error():
    """Error if minimum isn't numeric."""
    with pytest.raises(TypeError, match="Parameter minimum must be numeric"):
        Parameter(minimum="foo")  # type:ignore[arg-type]


def test_parameter_non_negative():

    notnonneg = Parameter(value=1, non_negative=False)
    valuenotnoneg, _, _ = notnonneg.get_value_and_bounds_for_optimization()
    assert np.allclose(1, valuenotnoneg)
    notnonneg.set_value_from_optimization(valuenotnoneg)
    assert np.allclose(1, notnonneg.value)

    nonneg1 = Parameter(value=1, non_negative=True)
    value1, _, _ = nonneg1.get_value_and_bounds_for_optimization()
    assert not np.allclose(1, value1)
    nonneg1.set_value_from_optimization(value1)
    assert np.allclose(1, nonneg1.value)

    nonneg2 = Parameter(value=2, non_negative=True)
    value2, _, _ = nonneg2.get_value_and_bounds_for_optimization()
    assert not np.allclose(2, value2)
    nonneg2.set_value_from_optimization(value2)
    assert np.allclose(2, nonneg2.value)

    nonnegminmax = Parameter(value=5, minimum=3, maximum=6, non_negative=True)
    value5, minimum, maximum = nonnegminmax.get_value_and_bounds_for_optimization()
    assert not np.allclose(5, value5)
    assert not np.allclose(3, minimum)
    assert not np.allclose(6, maximum)


def test_parameter_group_to_array():
    params = """
    - ["1", 1, {non-negative: false, min: -1, max: 1, vary: false}]
    - ["2", 4e2, {non-negative: true, min: 10, max: 8e2, vary: true}]
    - ["3", 2e4]
    """

    params = load_parameters(params, format_name="yml_str")

    labels, values, lower_bounds, upper_bounds = params.get_label_value_and_bounds_arrays(
        exclude_non_vary=False
    )

    assert len(labels) == 3
    assert len(values) == 3
    assert len(lower_bounds) == 3
    assert len(upper_bounds) == 3

    assert labels == ["1", "2", "3"]
    assert np.allclose(values, [1, np.log(4e2), 2e4])
    assert np.allclose(lower_bounds, [-1, np.log(10), -np.inf])
    assert np.allclose(upper_bounds, [1, np.log(8e2), np.inf])

    (
        labels_only_vary,
        values_only_vary,
        lower_bounds_only_vary,
        upper_bounds_only_vary,
    ) = params.get_label_value_and_bounds_arrays(exclude_non_vary=True)

    assert len(labels_only_vary) == 2
    assert len(values_only_vary) == 2
    assert len(lower_bounds_only_vary) == 2
    assert len(upper_bounds_only_vary) == 2

    assert labels_only_vary == ["2", "3"]


def test_update_parameter_group_from_array():
    params = """
    - ["1", 1, {non-negative: false, min: -1, max: 1, vary: false}]
    - ["2", 4e2, {non-negative: true, min: 10, max: 8e2, vary: true}]
    - ["3", 2e4]
    """

    params = load_parameters(params, format_name="yml_str")

    labels = ["1", "2", "3"]
    values = [0, np.log(6e2), 42]

    params.set_from_label_and_value_arrays(labels, values)

    values[1] = np.exp(values[1])

    for i in range(3):
        assert params.get(f"{i+1}").value == values[i]


@pytest.mark.parametrize(
    "case",
    [
        ("$1", "group.get('1').value"),
        (
            "1 - $kinetic.1 * exp($kinetic.2) + $kinetic.3",
            "1 - group.get('kinetic.1').value * exp(group.get('kinetic.2').value) "
            "+ group.get('kinetic.3').value",
        ),
        ("2", "2"),
        (
            "1 - sum([$kinetic.1, $kinetic.2])",
            "1 - sum([group.get('kinetic.1').value, group.get('kinetic.2').value])",
        ),
        ("exp($kinetic.4)", "exp(group.get('kinetic.4').value)"),
        ("$kinetic.5", "group.get('kinetic.5').value"),
        (
            "$group.sub_group.param1 + $kinetic6",
            "group.get('group.sub_group.param1').value + group.get('kinetic6').value",
        ),
        ("$foo.7.bar + $kinetic6", "group.get('foo.7.bar').value + group.get('kinetic6').value"),
        ("$1", "group.get('1').value"),
        ("$1-$2", "group.get('1').value-group.get('2').value"),
        ("$1-$5", "group.get('1').value-group.get('5').value"),
        (
            "100 - $inputs1.s1 - $inputs1.s3 - $inputs1.s8 - $inputs1.s12",
            "100 - group.get('inputs1.s1').value - group.get('inputs1.s3').value "
            "- group.get('inputs1.s8').value - group.get('inputs1.s12').value",
        ),
    ],
)
def test_transform_expression(case):
    expression, wanted_parameters = case
    parameter = Parameter(expression=expression)
    assert parameter.transformed_expression == wanted_parameters
    # just for the test coverage so the if condition is wrong
    assert parameter.transformed_expression == wanted_parameters


def test_label_validator():
    valid_names = [
        "1",
        "valid1",
        "_valid2",
        "extra_valid3",
    ]

    assert all(list(map(Parameter.valid_label, valid_names)))

    invalid_names = [
        "testé",
        "kinetic.1",
        "kinetic_red.3",
        "foo.7.bar",
        "_ilikeunderscoresatbegeninngin.justbecause",
        "42istheanswer.42",
        "kinetic::red",
        "kinetic_blue+kinetic_red",
        "makesthissense=trueandfalse",
        "what/about\\slashes",
        "$invalid",
        "round",
        "group",
    ]
    assert not any(list(map(Parameter.valid_label, invalid_names)))


def test_parameter_expressions():
    params = """
    - ["1", 2]
    - ["2", 5]
    - ["3", {expr: '$1 * exp($2)'}]
    - ["4", {expr: '2'}]
    """

    params = load_parameters(params, format_name="yml_str")

    assert params.get("3").expression is not None
    assert not params.get("3").vary
    assert params.get("3").value == 2 * np.exp(5)
    assert params.get("3").value == params.get("1") * np.exp(params.get("2"))
    assert params.get("4").value == 2

    with pytest.raises(ValueError):
        params_bad_expr = """
    - ["3", {expr: 'None'}]
    """
        load_parameters(params_bad_expr, format_name="yml_str")


def test_parameter_expressions_groups():
    params_vary_explicit = """
    b:
        - [0.25, {vary: True}]
        - [0.75, {expr: '1 - $b.1', vary: False}]
    rates:
        - ["total", 2, {vary: True}]
        - ["branch1", {expr: '$rates.total * $b.1', vary: False}]
        - ["branch2", {expr: '$rates.total * $b.2', vary: False}]
    """
    params_vary_implicit = """
    b:
        - [0.25]
        - [0.75, {expr: '1 - $b.1'}]
    rates:
        - ["total", 2]
        - ["branch1", {expr: '$rates.total * $b.1'}]
        - ["branch2", {expr: '$rates.total * $b.2'}]
    """
    params_label_explicit = """
    b:
        - ["1", 0.25]
        - ["2", 0.75, {expr: '1 - $b.1'}]
    rates:
        - ["total", 2]
        - ["branch1", {expr: '$rates.total * $b.1'}]
        - ["branch2", {expr: '$rates.total * $b.2'}]
    """

    for params in [params_vary_explicit, params_vary_implicit, params_label_explicit]:
        params = load_parameters(params, format_name="yml_str")

        assert params.get("b.1").expression is None
        assert params.get("b.1").vary
        assert not params.get("b.2").vary
        assert params.get("b.2").expression is not None
        assert params.get("rates.branch1").value == params.get("rates.total") * params.get("b.1")
        assert params.get("rates.branch2").value == params.get("rates.total") * params.get("b.2")
        assert params.get("rates.total").vary
        assert not params.get("rates.branch1").vary
        assert not params.get("rates.branch2").vary


def test_parameter_pickle(tmpdir):

    parameter = Parameter("testlabel", "testlabelfull", "testexpression", 1, 2, True, 42, False)

    with open(tmpdir.join("test_param_pickle"), "wb") as f:
        pickle.dump(parameter, f)
    with open(tmpdir.join("test_param_pickle"), "rb") as f:
        pickled_parameter = pickle.load(f)

    assert parameter == pickled_parameter


def test_parameter_numpy_operations():
    """Operators work like a float"""
    parm1 = Parameter(value=1)
    parm1_neg = Parameter(value=-1)
    parm2 = Parameter(value=2)
    parm3 = Parameter(value=3)
    parm3_5 = Parameter(value=3.5)

    assert parm1 == 1
    assert parm1 != 2
    assert -parm1 == -1
    assert abs(parm1_neg) == 1
    assert int(parm1) == 1
    assert np.allclose(float(parm1), 1)
    assert np.allclose(parm1 + parm2, 3)
    assert np.allclose(parm3 - parm2, 1)
    assert np.allclose(parm3 * parm2, 6)
    assert np.allclose(parm3 / parm2, 1.5)
    assert parm3 // parm2 == 1
    assert np.trunc(parm3_5) == 3
    assert parm2 % 2 == 0
    assert 2 % parm2 == 0
    assert divmod(parm3, 2) == (1, 1)
    assert divmod(3, parm2) == (1, 1)
    assert np.allclose(parm3 ** parm2, 9)
    assert parm3 > parm2
    assert parm3 >= parm2
    assert parm1 < parm2
    assert parm1 <= parm2


TEST_CSV = """
label, value, minimum, maximum, vary, non-negative, expression
rates.k1,0.050,0,5,True,True,None
rates.k2,None,-inf,inf,True,True,$rates.k1 * 2
rates.k3,2.311,-inf,inf,True,True,None
pen.eq.1,1.000,-inf,inf,False,False,None
"""


def test_param_group_from_csv(tmpdir):

    csv_path = tmpdir.join("parameters.csv")
    with open(csv_path, "w") as f:
        f.write(TEST_CSV)

    params = load_parameters(csv_path)

    assert "rates" in params

    assert params.has("rates.k1")
    p = params.get("rates.k1")
    assert p.label == "k1"
    assert p.value == 0.05
    assert p.minimum == 0
    assert p.maximum == 5
    assert p.vary
    assert p.non_negative
    assert p.expression is None

    assert params.has("rates.k2")
    p = params.get("rates.k2")
    assert p.label == "k2"
    assert p.value == params.get("rates.k1") * 2
    assert p.minimum == -np.inf
    assert p.maximum == np.inf
    assert not p.vary
    assert not p.non_negative
    assert p.expression == "$rates.k1 * 2"

    assert params.has("rates.k3")
    p = params.get("rates.k3")
    assert p.label == "k3"
    assert p.value == 2.311
    assert p.minimum == -np.inf
    assert p.maximum == np.inf
    assert p.vary
    assert p.non_negative
    assert p.expression is None

    assert "pen" in params
    assert "eq" in params["pen"]

    assert params.has("pen.eq.1")
    p = params.get("pen.eq.1")
    assert p.label == "1"
    assert p.value == 1.0
    assert p.minimum == -np.inf
    assert p.maximum == np.inf
    assert not p.vary
    assert not p.non_negative
    assert p.expression is None


def test_parameter_to_csv(tmpdir):
    csv_path = tmpdir.join("parameters.csv")
    params = load_parameters(
        """
    b:
        - ["1", 0.25, {vary: false, min: 0, max: 8}]
        - ["2", 0.75, {expr: '1 - $b.1', non-negative: true}]
    rates:
        - ["total", 2]
        - ["branch1", {expr: '$rates.total * $b.1'}]
        - ["branch2", {expr: '$rates.total * $b.2'}]
    """,
        format_name="yml_str",
    )

    save_parameters(params, csv_path, "csv")

    with open(csv_path) as f:
        print(f.read())
    params_from_csv = load_parameters(csv_path)

    for label, p in params.all():
        assert params_from_csv.has(label)
        p_from_csv = params_from_csv.get(label)
        assert p.label == p_from_csv.label
        assert p.value == p_from_csv.value
        assert p.minimum == p_from_csv.minimum
        assert p.maximum == p_from_csv.maximum
        assert p.vary == p_from_csv.vary
        assert p.non_negative == p_from_csv.non_negative
        assert p.expression == p_from_csv.expression



from __future__ import annotations

import numpy as np
import xarray as xr

from glotaran.model import DatasetModel
from glotaran.model import Megacomplex
from glotaran.model import megacomplex


@megacomplex(unique=True, register_as="baseline")
class BaselineMegacomplex(Megacomplex):
    def calculate_matrix(
        self,
        dataset_model: DatasetModel,
        indices: dict[str, int],
        **kwargs,
    ):
        model_axis = dataset_model.get_model_axis()
        clp_label = [f"{dataset_model.label}_baseline"]
        matrix = np.ones((model_axis.size, 1), dtype=np.float64)
        return clp_label, matrix

    def index_dependent(self, dataset_model: DatasetModel) -> bool:
        return False

    def finalize_data(
        self,
        dataset_model: DatasetModel,
        dataset: xr.Dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        if not is_full_model:
            dataset["baseline"] = dataset.clp.sel(clp_label=f"{dataset_model.label}_baseline")



from __future__ import annotations

import itertools
from typing import Any
from typing import NamedTuple

import numba as nb
import numpy as np
import xarray as xr

from glotaran.model import DatasetModel
from glotaran.model import Model
from glotaran.parameter import ParameterGroup


class CalculatedMatrix(NamedTuple):
    clp_labels: list[str]
    matrix: np.ndarray


def find_overlap(a, b, rtol=1e-05, atol=1e-08):
    ovr_a = []
    ovr_b = []
    start_b = 0
    for i, ai in enumerate(a):
        for j, bj in itertools.islice(enumerate(b), start_b, None):
            if np.isclose(ai, bj, rtol=rtol, atol=atol, equal_nan=False):
                ovr_a.append(i)
                ovr_b.append(j)
            elif bj > ai:  # (more than tolerance)
                break  # all the rest will be farther away
            else:  # bj < ai (more than tolerance)
                start_b += 1  # ignore further tests of this item
    return (ovr_a, ovr_b)


def find_closest_index(index: float, axis: np.ndarray):
    return np.abs(axis - index).argmin()


def get_min_max_from_interval(interval, axis):
    minimum = np.abs(axis.values - interval[0]).argmin() if not np.isinf(interval[0]) else 0
    maximum = (
        np.abs(axis.values - interval[1]).argmin() + 1 if not np.isinf(interval[1]) else axis.size
    )
    return slice(minimum, maximum)


def calculate_matrix(
    dataset_model: DatasetModel,
    indices: dict[str, int],
    as_global_model: bool = False,
) -> CalculatedMatrix:

    clp_labels = None
    matrix = None

    megacomplex_iterator = dataset_model.iterate_megacomplexes

    if as_global_model:
        megacomplex_iterator = dataset_model.iterate_global_megacomplexes
        dataset_model.swap_dimensions()

    for scale, megacomplex in megacomplex_iterator():
        this_clp_labels, this_matrix = megacomplex.calculate_matrix(dataset_model, indices)

        if scale is not None:
            this_matrix *= scale

        if matrix is None:
            clp_labels = this_clp_labels
            matrix = this_matrix
        else:
            clp_labels, matrix = combine_matrix(matrix, this_matrix, clp_labels, this_clp_labels)

    if as_global_model:
        dataset_model.swap_dimensions()

    return CalculatedMatrix(clp_labels, matrix)


def combine_matrix(matrix, this_matrix, clp_labels, this_clp_labels):
    tmp_clp_labels = clp_labels + [c for c in this_clp_labels if c not in clp_labels]
    tmp_matrix = np.zeros((matrix.shape[0], len(tmp_clp_labels)), dtype=np.float64)
    for idx, label in enumerate(tmp_clp_labels):
        if label in clp_labels:
            tmp_matrix[:, idx] += matrix[:, clp_labels.index(label)]
        if label in this_clp_labels:
            tmp_matrix[:, idx] += this_matrix[:, this_clp_labels.index(label)]
    return tmp_clp_labels, tmp_matrix


@nb.jit(nopython=True, parallel=True)
def apply_weight(matrix, weight):
    for i in nb.prange(matrix.shape[1]):
        matrix[:, i] *= weight


def reduce_matrix(
    matrix: CalculatedMatrix,
    model: Model,
    parameters: ParameterGroup,
    index: Any | None,
) -> CalculatedMatrix:
    matrix = apply_relations(matrix, model, parameters, index)
    matrix = apply_constraints(matrix, model, index)
    return matrix


def apply_constraints(
    matrix: CalculatedMatrix,
    model: Model,
    index: Any | None,
) -> CalculatedMatrix:

    if len(model.clp_constraints) == 0:
        return matrix

    clp_labels = matrix.clp_labels
    removed_clp_labels = [
        c.target for c in model.clp_constraints if c.target in clp_labels and c.applies(index)
    ]
    reduced_clp_labels = [c for c in clp_labels if c not in removed_clp_labels]
    mask = [label in reduced_clp_labels for label in clp_labels]
    reduced_matrix = matrix.matrix[:, mask]
    return CalculatedMatrix(reduced_clp_labels, reduced_matrix)


def apply_relations(
    matrix: CalculatedMatrix,
    model: Model,
    parameters: ParameterGroup,
    index: Any | None,
) -> CalculatedMatrix:

    if len(model.clp_relations) == 0:
        return matrix

    clp_labels = matrix.clp_labels
    relation_matrix = np.diagflat([1.0 for _ in clp_labels])

    idx_to_delete = []
    for relation in model.clp_relations:
        if relation.target in clp_labels and relation.applies(index):

            if relation.source not in clp_labels:
                continue

            relation = relation.fill(model, parameters)
            source_idx = clp_labels.index(relation.source)
            target_idx = clp_labels.index(relation.target)
            relation_matrix[target_idx, source_idx] = relation.parameter
            idx_to_delete.append(target_idx)

    reduced_clp_labels = [label for i, label in enumerate(clp_labels) if i not in idx_to_delete]
    relation_matrix = np.delete(relation_matrix, idx_to_delete, axis=1)
    reduced_matrix = matrix.matrix @ relation_matrix
    return CalculatedMatrix(reduced_clp_labels, reduced_matrix)


def retrieve_clps(
    model: Model,
    parameters: ParameterGroup,
    clp_labels: xr.DataArray,
    reduced_clp_labels: xr.DataArray,
    reduced_clps: xr.DataArray,
    index: Any | None,
) -> xr.DataArray:
    if len(model.clp_relations) == 0 and len(model.clp_constraints) == 0:
        return reduced_clps

    clps = np.zeros(len(clp_labels))

    for i, label in enumerate(reduced_clp_labels):
        idx = clp_labels.index(label)
        clps[idx] = reduced_clps[i]

    for relation in model.clp_relations:
        relation = relation.fill(model, parameters)
        if (
            relation.target in clp_labels
            and relation.applies(index)
            and relation.source in clp_labels
        ):
            source_idx = clp_labels.index(relation.source)
            target_idx = clp_labels.index(relation.target)
            clps[target_idx] = relation.parameter * clps[source_idx]
    return clps


def calculate_clp_penalties(
    model: Model,
    parameters: ParameterGroup,
    clp_labels: list[list[str]] | list[str],
    clps: list[np.ndarray],
    global_axis: np.ndarray,
    dataset_models: dict[str, DatasetModel],
) -> np.ndarray:

    # TODO: make a decision on how to handle clp_penalties per dataset
    # 1. sum up contributions per dataset on each dataset_axis (v0.4.1)
    # 2. sum up contributions on the global_axis (future?)

    penalties = []
    for penalty in model.clp_area_penalties:
        penalty = penalty.fill(model, parameters)
        source_area = np.array([])
        target_area = np.array([])
        for _, dataset_model in dataset_models.items():
            dataset_axis = dataset_model.get_global_axis()

            source_area = np.concatenate(
                [
                    source_area,
                    _get_area(
                        penalty.source,
                        clp_labels,
                        clps,
                        penalty.source_intervals,
                        global_axis,
                        dataset_axis,
                    ),
                ]
            )

            target_area = np.concatenate(
                [
                    target_area,
                    _get_area(
                        penalty.target,
                        clp_labels,
                        clps,
                        penalty.target_intervals,
                        global_axis,
                        dataset_axis,
                    ),
                ]
            )
        area_penalty = np.abs(np.sum(source_area) - penalty.parameter * np.sum(target_area))

        penalties.append(area_penalty * penalty.weight)

    return np.asarray(penalties)


def _get_area(
    clp_label: str,
    clp_labels: list[list[str]],
    clps: list[np.ndarray],
    intervals: list[tuple[float, float]],
    global_axis: np.ndarray,
    dataset_axis: np.ndarray,
) -> np.ndarray:
    area = []

    for interval in intervals:
        if interval[0] > global_axis[-1]:
            continue
        bounded_interval = (
            max(interval[0], np.min(dataset_axis)),
            min(interval[1], np.max(dataset_axis)),
        )
        start_idx, end_idx = get_idx_from_interval(bounded_interval, global_axis)
        for i in range(start_idx, end_idx + 1):
            index_clp_labels = clp_labels[i] if isinstance(clp_labels[0], list) else clp_labels
            if clp_label in index_clp_labels:
                area.append(clps[i][index_clp_labels.index(clp_label)])

    return np.asarray(area)  # TODO: normalize for distance on global axis


def get_idx_from_interval(interval: tuple[float, float], axis: np.ndarray) -> tuple[int, int]:
    """Retrieves start and end index of an interval on some axis
    Parameters
    ----------
    interval : A tuple of floats with begin and end of the interval
    axis : Array like object which can be cast to np.array
    Returns
    -------
    start, end : tuple of int
    """
    start = np.abs(axis - interval[0]).argmin() if not np.isinf(interval[0]) else 0
    end = np.abs(axis - interval[1]).argmin() if not np.isinf(interval[1]) else axis.size - 1
    return start, end



import numpy as np
import pytest
import xarray as xr

from glotaran.analysis.optimize import optimize
from glotaran.analysis.simulation import simulate
from glotaran.analysis.util import calculate_matrix
from glotaran.builtin.megacomplexes.coherent_artifact import CoherentArtifactMegacomplex
from glotaran.builtin.megacomplexes.decay import DecayMegacomplex
from glotaran.model import Model
from glotaran.parameter import ParameterGroup
from glotaran.project import Scheme


@pytest.mark.parametrize(
    "is_index_dependent",
    (True, False),
)
def test_coherent_artifact(is_index_dependent: bool):
    model_dict = {
        "initial_concentration": {
            "j1": {"compartments": ["s1"], "parameters": ["irf_center"]},
        },
        "megacomplex": {
            "mc1": {"type": "decay", "k_matrix": ["k1"]},
            "mc2": {"type": "coherent-artifact", "order": 3},
        },
        "k_matrix": {
            "k1": {
                "matrix": {
                    ("s1", "s1"): "rate",
                }
            }
        },
        "irf": {
            "irf1": {
                "type": "spectral-multi-gaussian",
                "center": ["irf_center"],
                "width": ["irf_width"],
            },
        },
        "dataset": {
            "dataset1": {
                "initial_concentration": "j1",
                "megacomplex": ["mc1", "mc2"],
                "irf": "irf1",
            },
        },
    }

    parameter_list = [
        ["rate", 101e-4],
        ["irf_center", 10, {"vary": False, "non-negative": False}],
        ["irf_width", 20, {"vary": False, "non-negative": False}],
    ]

    if is_index_dependent is True:
        irf_spec = model_dict["irf"]["irf1"]
        irf_spec["dispersion_center"] = "irf_dispc"
        irf_spec["center_dispersion"] = ["irf_disp1", "irf_disp2"]

        parameter_list += [
            ["irf_dispc", 300, {"vary": False, "non-negative": False}],
            ["irf_disp1", 0.01, {"vary": False, "non-negative": False}],
            ["irf_disp2", 0.001, {"vary": False, "non-negative": False}],
        ]

    model = Model.from_dict(
        model_dict.copy(),
        megacomplex_types={
            "decay": DecayMegacomplex,
            "coherent-artifact": CoherentArtifactMegacomplex,
        },
    )

    parameters = ParameterGroup.from_list(parameter_list)

    time = np.arange(0, 50, 1.5)
    spectral = np.asarray([200, 300, 400])
    coords = {"time": time, "spectral": spectral}

    dataset_model = model.dataset["dataset1"].fill(model, parameters)
    dataset_model.overwrite_global_dimension("spectral")
    dataset_model.set_coordinates(coords)
    matrix = calculate_matrix(dataset_model, {"spectral": [0, 1, 2]})
    compartments = matrix.clp_labels

    print(compartments)
    assert len(compartments) == 4
    for i in range(1, 4):
        assert compartments[i] == f"coherent_artifact_{i}"

    assert matrix.matrix.shape == (time.size, 4)

    clp = xr.DataArray(
        np.ones((3, 4)),
        coords=[
            ("spectral", spectral),
            (
                "clp_label",
                [
                    "s1",
                    "coherent_artifact_1",
                    "coherent_artifact_2",
                    "coherent_artifact_3",
                ],
            ),
        ],
    )
    axis = {"time": time, "spectral": clp.spectral}
    data = simulate(model, "dataset1", parameters, axis, clp)

    dataset = {"dataset1": data}
    scheme = Scheme(
        model=model, parameters=parameters, data=dataset, maximum_number_function_evaluations=20
    )
    result = optimize(scheme)
    print(result.optimized_parameters)

    for label, param in result.optimized_parameters.all():
        assert np.allclose(param.value, parameters.get(label).value, rtol=1e-8)

    resultdata = result.data["dataset1"]
    assert np.array_equal(data.time, resultdata.time)
    assert np.array_equal(data.spectral, resultdata.spectral)
    assert data.data.shape == resultdata.data.shape
    assert data.data.shape == resultdata.fitted_data.shape
    assert np.allclose(data.data, resultdata.fitted_data)

    assert "coherent_artifact_response" in resultdata
    if is_index_dependent:
        assert resultdata["coherent_artifact_response"].shape == (spectral.size, time.size, 3)
    else:
        assert resultdata["coherent_artifact_response"].shape == (time.size, 3)

    assert "coherent_artifact_associated_spectra" in resultdata
    assert resultdata["coherent_artifact_associated_spectra"].shape == (3, 3)



from __future__ import annotations

import pandas as pd

from glotaran.io import ProjectIoInterface
from glotaran.io import register_project_io
from glotaran.parameter import ParameterGroup


@register_project_io(["csv"])
class CsvProjectIo(ProjectIoInterface):
    def load_parameters(self, file_name: str) -> ParameterGroup:
        df = pd.read_csv(file_name, skipinitialspace=True, na_values=["None", "none"])
        return ParameterGroup.from_dataframe(df, source=file_name)

    def save_parameters(self, parameters: ParameterGroup, file_name: str, *, sep=","):
        """Save a :class:`ParameterGroup` to a CSV file."""
        parameters.to_dataframe().to_csv(file_name, na_rep="None", index=False, sep=sep)



"""Contains helper methods for dataclasses."""
from __future__ import annotations

import dataclasses
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any
    from typing import Callable
    from typing import TypeVar

    DefaultType = TypeVar("DefaultType")


def exclude_from_dict_field(
    default: DefaultType = dataclasses.MISSING,  # type:ignore[assignment]
) -> DefaultType:
    """Create a dataclass field with which will be excluded from ``asdict``.

    Parameters
    ----------
    default : DefaultType
        The default value of the field.

    Returns
    -------
    DefaultType
        The created field.
    """
    return dataclasses.field(default=default, metadata={"exclude_from_dict": True})


def file_representation_field(
    target: str,
    loader: Callable[[str], Any],
    default: DefaultType = dataclasses.MISSING,  # type:ignore[assignment]
) -> DefaultType:
    """Create a dataclass field with target and loader as metadata.

    Parameters
    ----------
    target : str
        The name of the represented field.
    loader : Callable[[str], Any]
        A function to load the target field from a file.
    default : DefaultType
        The default value of the field.

    Returns
    -------
    DefaultType
        The created field.
    """
    return dataclasses.field(default=default, metadata={"target": target, "loader": loader})


def asdict(dataclass: object) -> dict[str, Any]:
    """Create a dictionary containing all fields of the dataclass.

    Parameters
    ----------
    dataclass : object
        A dataclass instance.

    Returns
    -------
    dict[str, Any] :
        The dataclass represented as a dictionary.
    """
    fields = dataclasses.fields(dataclass)

    dataclass_dict = {}
    for field in fields:
        if "exclude_from_dict" not in field.metadata:
            value = getattr(dataclass, field.name)
            dataclass_dict[field.name] = (
                asdict(value) if dataclasses.is_dataclass(value) else value
            )

    return dataclass_dict


def fromdict(dataclass_type: type, dataclass_dict: dict[str, Any], folder: Path = None) -> object:
    """Create a dataclass instance from a dict and loads all file represented fields.

    Parameters
    ----------
    dataclass_type : type
        A dataclass type.
    dataclass_dict : dict[str, Any]
        A dict for instancing the the dataclass.
    folder : Path
        The root folder for file paths. If ``None`` file paths are consider absolute.

    Returns
    -------
    object
        Created instance of dataclass_type.
    """
    fields = dataclasses.fields(dataclass_type)

    for field in fields:
        if "target" in field.metadata and "loader" in field.metadata:
            file_path = dataclass_dict.get(field.name)
            if file_path is None:
                continue
            elif isinstance(file_path, list):
                dataclass_dict[field.metadata["target"]] = [
                    field.metadata["loader"](f if folder is None else folder / f)
                    for f in file_path
                ]
            elif isinstance(file_path, dict):
                dataclass_dict[field.metadata["target"]] = {
                    k: field.metadata["loader"](f if folder is None else folder / f)
                    for k, f in file_path.items()
                }
            else:
                dataclass_dict[field.metadata["target"]] = field.metadata["loader"](
                    file_path if folder is None else folder / file_path
                )
        elif dataclasses.is_dataclass(field.default) and field.name in dataclass_dict:
            dataclass_dict[field.name] = type(field.default)(**dataclass_dict[field.name])

    return dataclass_type(**dataclass_dict)



from __future__ import annotations

from typing import TYPE_CHECKING

from glotaran.examples.sequential import model
from glotaran.io import load_model
from glotaran.io import save_model

if TYPE_CHECKING:
    from pathlib import Path


want = """dataset:
  dataset1:
    group: default
    initial_concentration: j1
    irf: irf1
    megacomplex:
    - m1
dataset_groups:
  default:
    link_clp: null
    residual_function: variable_projection
default_megacomplex: decay
initial_concentration:
  j1:
    compartments:
    - s1
    - s2
    - s3
    exclude_from_normalize: []
    parameters:
    - j.1
    - j.0
    - j.0
irf:
  irf1:
    backsweep: false
    center: irf.center
    normalize: true
    type: gaussian
    width: irf.width
k_matrix:
  k1:
    matrix:
      (s2, s1): kinetic.1
      (s3, s2): kinetic.2
      (s3, s3): kinetic.3
megacomplex:
  m1:
    dimension: time
    k_matrix:
    - k1
    type: decay
"""


def test_save_model(
    tmp_path: Path,
):
    """Check all files exist."""

    model_path = tmp_path / "testmodel.yml"
    save_model(file_name=model_path, format_name="yml", model=model)

    assert model_path.is_file()
    assert model_path.read_text() == want
    assert load_model(model_path).valid()



from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

from glotaran.analysis.optimize import optimize
from glotaran.analysis.simulation import simulate
from glotaran.model import Model
from glotaran.parameter import ParameterGroup
from glotaran.project import Scheme
from glotaran.testing.model_generators import SimpleModelGenerator


def _create_gaussian_clp(labels, amplitudes, centers, widths, axis):
    return xr.DataArray(
        [
            amplitudes[i] * np.exp(-np.log(2) * np.square(2 * (axis - centers[i]) / widths[i]))
            for i, _ in enumerate(labels)
        ],
        coords=[("clp_label", labels), ("pixel", axis.data)],
    ).T


class DecayModel(Model):
    @classmethod
    def from_dict(
        cls,
        model_dict,
    ):
        model_dict = {**model_dict, "default_megacomplex": "decay"}
        return super().from_dict(model_dict)


class OneComponentOneChannel:
    model = DecayModel.from_dict(
        {
            "initial_concentration": {
                "j1": {"compartments": ["s1"], "parameters": ["2"]},
            },
            "megacomplex": {
                "mc1": {"k_matrix": ["k1"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s1", "s1"): "1",
                    }
                }
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    initial_parameters = ParameterGroup.from_list(
        [101e-4, [1, {"vary": False, "non-negative": False}]]
    )
    wanted_parameters = ParameterGroup.from_list(
        [101e-3, [1, {"vary": False, "non-negative": False}]]
    )

    time = np.arange(0, 50, 1.5)
    pixel = np.asarray([0])
    axis = {"time": time, "pixel": pixel}

    clp = xr.DataArray([[1]], coords=[("pixel", pixel.data), ("clp_label", ["s1"])])


class OneComponentOneChannelGaussianIrf:
    model = DecayModel.from_dict(
        {
            "initial_concentration": {
                "j1": {"compartments": ["s1"], "parameters": ["5"]},
            },
            "megacomplex": {
                "mc1": {"k_matrix": ["k1"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s1", "s1"): "1",
                    }
                }
            },
            "irf": {
                "irf1": {"type": "gaussian", "center": "2", "width": "3", "shift": ["4"]},
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "irf": "irf1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    initial_parameters = ParameterGroup.from_list(
        [101e-4, 0.1, 1, [0.1, {"vary": False}], [1, {"vary": False, "non-negative": False}]]
    )
    assert model.megacomplex["mc1"].index_dependent(
        model.dataset["dataset1"].fill(model, initial_parameters)
    )
    wanted_parameters = ParameterGroup.from_list(
        [
            [101e-3, {"non-negative": True}],
            [0.2, {"non-negative": True}],
            [2, {"non-negative": True}],
            [0.1, {"vary": False}],
            [1, {"vary": False, "non-negative": False}],
        ]
    )

    time = np.arange(0, 50, 1.5)
    pixel = np.asarray([0])
    axis = {"time": time, "pixel": pixel}

    clp = xr.DataArray([[1]], coords=[("pixel", pixel.data), ("clp_label", ["s1"])])


class ThreeComponentParallel:
    generator = SimpleModelGenerator(
        rates=[300e-3, 500e-4, 700e-5],
        irf={"center": 1.3, "width": 7.8},
        k_matrix="parallel",
    )
    model, initial_parameters = generator.model_and_parameters

    generator.rates = [301e-3, 502e-4, 705e-5]
    wanted_parameters = generator.parameters

    time = np.arange(-10, 100, 1.5)
    pixel = np.arange(600, 750, 10)

    axis = {"time": time, "pixel": pixel}

    clp = _create_gaussian_clp(
        ["s1", "s2", "s3"], [7, 3, 30], [620, 670, 720], [10, 30, 50], pixel
    )


class ThreeComponentSequential:
    model = DecayModel.from_dict(
        {
            "initial_concentration": {
                "j1": {"compartments": ["s1", "s2", "s3"], "parameters": ["j.1", "j.0", "j.0"]},
            },
            "megacomplex": {
                "mc1": {"k_matrix": ["k1"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s2", "s1"): "kinetic.1",
                        ("s3", "s2"): "kinetic.2",
                        ("s3", "s3"): "kinetic.3",
                    }
                }
            },
            "irf": {
                "irf1": {
                    "type": "multi-gaussian",
                    "center": ["irf.center"],
                    "width": ["irf.width"],
                },
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "irf": "irf1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    initial_parameters = ParameterGroup.from_dict(
        {
            "kinetic": [
                ["1", 501e-3],
                ["2", 202e-4],
                ["3", 105e-5],
                {"non-negative": True},
            ],
            "irf": [["center", 1.3], ["width", 7.8]],
            "j": [
                ["1", 1, {"vary": False, "non-negative": False}],
                ["0", 0, {"vary": False, "non-negative": False}],
            ],
        }
    )
    wanted_parameters = ParameterGroup.from_dict(
        {
            "kinetic": [
                ["1", 501e-3],
                ["2", 202e-4],
                ["3", 105e-5],
            ],
            "irf": [["center", 1.3], ["width", 7.8]],
            "j": [
                ["1", 1, {"vary": False, "non-negative": False}],
                ["0", 0, {"vary": False, "non-negative": False}],
            ],
        }
    )

    time = np.arange(-10, 50, 1.0)
    pixel = np.arange(600, 750, 10)
    axis = {"time": time, "pixel": pixel}

    clp = _create_gaussian_clp(
        ["s1", "s2", "s3"], [7, 3, 30], [620, 670, 720], [10, 30, 50], pixel
    )


@pytest.mark.parametrize(
    "suite",
    [
        OneComponentOneChannel,
        OneComponentOneChannelGaussianIrf,
        ThreeComponentParallel,
        ThreeComponentSequential,
    ],
)
@pytest.mark.parametrize("nnls", [True, False])
def test_kinetic_model(suite, nnls):

    model = suite.model
    print(model.validate())
    assert model.valid()
    model.dataset_group_models["default"].method = (
        "non_negative_least_squares" if nnls else "variable_projection"
    )

    wanted_parameters = suite.wanted_parameters
    print(model.validate(wanted_parameters))
    print(wanted_parameters)
    assert model.valid(wanted_parameters)

    initial_parameters = suite.initial_parameters
    print(model.validate(initial_parameters))
    assert model.valid(initial_parameters)

    print(model.markdown(initial_parameters))

    dataset = simulate(model, "dataset1", wanted_parameters, suite.axis, suite.clp)

    assert dataset.data.shape == (suite.axis["time"].size, suite.axis["pixel"].size)

    data = {"dataset1": dataset}

    scheme = Scheme(
        model=model,
        parameters=initial_parameters,
        data=data,
        maximum_number_function_evaluations=20,
    )
    result = optimize(scheme)
    print(result.optimized_parameters)

    for label, param in result.optimized_parameters.all():
        assert np.allclose(param.value, wanted_parameters.get(label).value, rtol=1e-1)

    resultdata = result.data["dataset1"]
    assert np.array_equal(dataset["time"], resultdata["time"])
    assert np.array_equal(dataset["pixel"], resultdata["pixel"])
    assert dataset.data.shape == resultdata.data.shape
    assert dataset.data.shape == resultdata.fitted_data.shape
    assert np.allclose(dataset.data, resultdata.fitted_data, rtol=1e-2)
    assert "species_associated_images" in resultdata
    assert "decay_associated_images" in resultdata

    if len(model.irf) != 0:
        assert "irf" in resultdata



"""Model generators used to generate simple models from a set of inputs."""

from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING
from typing import Literal

from glotaran.model import Model
from glotaran.parameter.parameter_group import ParameterGroup

if TYPE_CHECKING:
    from glotaran.utils.ipython import MarkdownStr


def _split_iterable_in_non_dict_and_dict_items(
    input_list: list[float, dict[str, bool | float]],
) -> tuple[list[float], list[dict[str, bool | float]]]:
    """Split an iterable (list) into non-dict and dict items.

    Parameters
    ----------
    input_list : list[float, dict[str, bool | float]]
        A list of values of type `float` and a dict with parameter options, e.g.
        `[1, 2, 3, {"vary": False, "non-negative": True}]`

    Returns
    -------
    tuple[list[float], list[dict[str, bool | float]]]
        Split a list into non-dict (`values`) and dict items (`defaults`),
        return a tuple (`values`, `defaults`)
    """
    values: list = [val for val in input_list if not isinstance(val, dict)]
    defaults: list = [val for val in input_list if isinstance(val, dict)]
    return values, defaults


@dataclass
class SimpleModelGenerator:
    """A minimal boilerplate model and parameters generator.

    Generates a model (together with the parameters specification) based on
    parameter input values assigned to the generator's attributes
    """

    rates: list[float] = field(default_factory=list)
    """A list of values representing decay rates"""
    k_matrix: Literal["parallel", "sequential"] | dict[tuple[str, str], str] = "parallel"
    """"A `dict` with a k_matrix specification or `Literal["parallel", "sequential"]`"""
    compartments: list[str] | None = None
    """A list of compartment names"""
    irf: dict[str, float] = field(default_factory=dict)
    """A dict of items specifying an irf"""
    initial_concentration: list[float] = field(default_factory=list)
    """A list values representing the initial concentration"""
    dispersion_coefficients: list[float] = field(default_factory=list)
    """A list of values representing the dispersion coefficients"""
    dispersion_center: float | None = None
    """A value representing the dispersion center"""
    default_megacomplex: str = "decay"
    """The default_megacomplex identifier"""
    # TODO: add support for a spectral model:
    # shapes: list[float] = field(default_factory=list, init=False)

    @property
    def valid(self) -> bool:
        """Check if the generator state is valid.

        Returns
        -------
        bool
            Generator state obtained by calling the generated model's
            `valid` function with the generated parameters as input.
        """
        try:
            return self.model.valid(parameters=self.parameters)
        except ValueError:
            return False

    def validate(self) -> str:
        """Call `validate` on the generated model and return its output.

        Returns
        -------
        str
            A string listing problems in the generated model and parameters if any.
        """
        return self.model.validate(parameters=self.parameters)

    @property
    def model(self) -> Model:
        """Return the generated model.

        Returns
        -------
        Model
            The generated model of type :class:`glotaran.model.Model`.
        """
        return Model.from_dict(self.model_dict)

    @property
    def model_dict(self) -> dict:
        """Return a dict representation of the generated model.

        Returns
        -------
        dict
            A dict representation of the generated model.
        """
        return self._model_dict()

    @property
    def parameters(self) -> ParameterGroup:
        """Return the generated parameters of type :class:`glotaran.parameter.ParameterGroup`.

        Returns
        -------
        ParameterGroup
            The generated parameters of type of type :class:`glotaran.parameter.ParameterGroup`.
        """
        return ParameterGroup.from_dict(self.parameters_dict)

    @property
    def parameters_dict(self) -> dict:
        """Return a dict representation of the generated parameters.

        Returns
        -------
        dict
            A dict representing the generated parameters.
        """
        return self._parameters_dict()

    @property
    def model_and_parameters(self) -> tuple[Model, ParameterGroup]:
        """Return generated model and parameters.

        Returns
        -------
        tuple[Model, ParameterGroup]
            A model of type :class:`glotaran.model.Model` and
            and parameters of type :class:`glotaran.parameter.ParameterGroup`.
        """
        return self.model, self.parameters

    @property
    def _rates(self) -> tuple[list[float], list[dict[str, bool | float]]]:
        """Validate input to rates, return a tuple of rates and parameter defaults.

        Returns
        -------
        tuple[list[float], list[dict[str, bool | float]]]
            A tuple of a list of rates and a dict containing parameter defaults

        Raises
        ------
        ValueError
            Raised if rates is not a list of at least one number.
        """
        if not isinstance(self.rates, list):
            raise ValueError(f"generator.rates: must be a `list`, got: {self.rates}")
        if len(self.rates) == 0:
            raise ValueError("generator.rates: must be a `list` with 1 or more rates")
        if not isinstance(self.rates[0], (int, float)):
            raise ValueError(f"generator.rates: 1st element must be numeric, got: {self.rates[0]}")
        return _split_iterable_in_non_dict_and_dict_items(self.rates)

    def _parameters_dict_items(self) -> dict:
        """Return a dict with items used in constructing the parameters.

        Returns
        -------
        dict
            A dict with items used in constructing a parameters dict.
        """
        rates, rates_defaults = self._rates
        items = {"rates": rates}
        if rates_defaults:
            items["rates_defaults"] = rates_defaults[0]
        items["irf"] = [[key, value] for key, value in self.irf.items()]
        if self.initial_concentration:
            items["inputs"] = self.initial_concentration
        elif self.k_matrix == "parallel":
            items["inputs"] = [
                ["1", 1],
                {"vary": False},
            ]
        elif self.k_matrix == "sequential":
            items["inputs"] = [
                ["1", 1],
                ["0", 0],
                {"vary": False},
            ]
        return items

    def _model_dict_items(self) -> dict:
        """Return a dict with items used in constructing the model.

        Returns
        -------
        dict
            A dict with items used in constructing a model dict.
        """
        rates, _ = self._rates
        nr = len(rates)
        indices = list(range(1, 1 + nr))
        items = {"default_megacomplex": self.default_megacomplex}
        if self.irf:
            items["irf"] = {
                "type": "multi-gaussian",
                "center": ["irf.center"],
                "width": ["irf.width"],
            }
        if isinstance(self.k_matrix, dict):
            items["k_matrix"] = self.k_matrix
            items["input_parameters"] = [f"inputs.{i}" for i in indices]
            items["compartments"] = [f"s{i}" for i in indices]
            # TODO: get unique compartments from user defined k_matrix
        if self.k_matrix == "parallel":
            items["input_parameters"] = ["inputs.1"] * nr
            items["k_matrix"] = {(f"s{i}", f"s{i}"): f"rates.{i}" for i in indices}
        elif self.k_matrix == "sequential":
            items["input_parameters"] = ["inputs.1"] + ["inputs.0"] * (nr - 1)
            items["k_matrix"] = {
                (f"s{i if i==nr else i+1}", f"s{i}"): f"rates.{i}" for i in indices
            }

        if self.k_matrix in ("parallel", "sequential"):
            items["compartments"] = [f"s{i}" for i in indices]
        return items

    def _parameters_dict(self) -> dict:
        """Return a parameters dict.

        Returns
        -------
        dict
            A dict that can be passed to the `ParameterGroup` `from_dict` method.
        """
        items = self._parameters_dict_items()
        rates = items["rates"]
        if "rates_defaults" in items:
            rates += [items["rates_defaults"]]
        result = {"rates": rates}
        if items["irf"]:
            result["irf"] = items["irf"]
        result["inputs"] = items["inputs"]
        return result

    def _model_dict(self) -> dict:
        """Return a model dict.

        Returns
        -------
        dict
            A dict that can be passed to the `Model` `from_dict` method.
        """
        items = self._model_dict_items()
        result = {"default_megacomplex": items["default_megacomplex"]}
        result.update(
            {
                "initial_concentration": {
                    "j1": {
                        "compartments": items["compartments"],
                        "parameters": items["input_parameters"],
                    },
                },
                "megacomplex": {
                    "mc1": {"k_matrix": ["k1"]},
                },
                "k_matrix": {"k1": {"matrix": items["k_matrix"]}},
                "dataset": {
                    "dataset1": {
                        "initial_concentration": "j1",
                        "megacomplex": ["mc1"],
                    },
                },
            }
        )
        if "irf" in items:
            result["dataset"]["dataset1"].update({"irf": "irf1"})
            result["irf"] = {
                "irf1": items["irf"],
            }
        return result

    def markdown(self) -> MarkdownStr:
        """Return a markdown string representation of the generated model and parameters.

        Returns
        -------
        MarkdownStr
            A markdown string
        """
        return self.model.markdown(parameters=self.parameters)



"""Tests for deprecated methods in ``glotaran..parameter.ParameterGroup``."""
from pathlib import Path
from textwrap import dedent

from glotaran.deprecation.modules.test import deprecation_warning_on_call_test_helper
from glotaran.examples.sequential import parameter


def test_parameter_group_to_csv(tmp_path: Path):
    """``ParameterGroup.to_csv`` raises deprecation warning and saves file."""
    parameter_path = tmp_path / "test_parameter.csv"
    deprecation_warning_on_call_test_helper(
        parameter.to_csv, args=[parameter_path.as_posix()], raise_exception=True
    )
    expected = dedent(
        """\
        label,value,minimum,maximum,vary,non-negative,expression
        j.1,1.0,-inf,inf,False,False,None
        j.0,0.0,-inf,inf,False,False,None
        kinetic.1,0.5,-inf,inf,True,False,None
        kinetic.2,0.3,-inf,inf,True,False,None
        kinetic.3,0.1,-inf,inf,True,False,None
        irf.center,0.3,-inf,inf,True,False,None
        irf.width,0.1,-inf,inf,True,False,None
        """
    )

    assert parameter_path.is_file()
    assert parameter_path.read_text() == expected



""" K-Matrix """
from __future__ import annotations

import itertools
import typing
from collections import OrderedDict

import numpy as np
from scipy.linalg import eig
from scipy.linalg import solve

from glotaran.builtin.megacomplexes.decay.initial_concentration import InitialConcentration
from glotaran.model import model_item
from glotaran.parameter import Parameter
from glotaran.utils.ipython import MarkdownStr


@model_item(
    properties={
        "matrix": {"type": typing.Dict[typing.Tuple[str, str], Parameter]},
    },
)
class KMatrix:
    """A K-Matrix represents a first order differental system."""

    @classmethod
    def empty(cls, label: str, compartments: list[str]) -> KMatrix:
        """Creates an empty K-Matrix. Useful for combining.

        Parameters
        ----------
        label :
            Label of the K-Matrix

        compartments :
            A list of all compartments in the model.
        """
        return cls(label, OrderedDict())

    def involved_compartments(self) -> list[str]:
        """A list of all compartments in the Matrix."""
        # TODO: find a better way that preserves ordering as defined in initial_concentrations
        compartments = []
        for index in self.matrix:
            if index[0] not in compartments:
                compartments.append(index[0])
            if index[1] not in compartments:
                compartments.append(index[1])

        # Don't use set, it randomly reorders the compartments.
        # compartments = list(set(compartments))
        return compartments

    def combine(self, k_matrix: KMatrix) -> KMatrix:
        """Creates a combined matrix.

        When combining k-matrices km1 and km2 (km1.combine(km2)),
        entries in km1 will be overwritten by corresponding entries in km2.

        Parameters
        ----------
        k_matrix :
            KMatrix to combine with.

        Returns
        -------
        combined :
            The combined KMatrix.

        """
        if not isinstance(k_matrix, KMatrix):
            raise TypeError("K-Matrices can only be combined with other K-Matrices.")
        combined_matrix = {entry: self.matrix[entry] for entry in self.matrix}
        for entry in k_matrix.matrix:
            combined_matrix[entry] = k_matrix.matrix[entry]
        combined = KMatrix()
        combined.label = f"{self.label}+{k_matrix.label}"
        combined.matrix = combined_matrix
        return combined

    def matrix_as_markdown(
        self,
        compartments: list[str] = None,
        fill_parameters: bool = False,
    ) -> MarkdownStr:
        """Returns the KMatrix as markdown formatted table.

        Parameters
        ----------
        compartments :
            (default = None)
            An optional list defining the desired order of compartments.
        fill_parameters : bool
            (default = False)
            If true, the entries will be filled with the actual parameter values
            instead of labels.
        """

        compartments = (
            [c for c in compartments if c in self.involved_compartments()]
            if compartments
            else self.involved_compartments()
        )
        size = len(compartments)
        array = np.zeros((size, size), dtype=object)
        # Matrix is a dict
        for index in self.matrix:
            i = compartments.index(index[0])
            j = compartments.index(index[1])
            array[i, j] = (
                self.matrix[index].full_label if not fill_parameters else self.matrix[index].value
            )
        return self._array_as_markdown(array, compartments, compartments)

    def _repr_markdown_(self) -> str:
        """Special method used by ``ipython`` to render markdown."""
        return str(self.matrix_as_markdown())

    def a_matrix_as_markdown(self, initial_concentration: InitialConcentration) -> MarkdownStr:
        """Returns the A Matrix as markdown formatted table.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        compartments = [
            c for c in initial_concentration.compartments if c in self.involved_compartments()
        ]
        return self._array_as_markdown(
            self.a_matrix(initial_concentration).T,
            compartments,
            self.rates(initial_concentration),
        )

    @staticmethod
    def _array_as_markdown(array, row_header, column_header) -> MarkdownStr:
        markdown = "| compartment | "
        markdown += " | ".join(f"{e:.4e}" if not isinstance(e, str) else e for e in column_header)

        markdown += "\n|"
        markdown += "|".join("---" for _ in range(len(column_header) + 1))
        markdown += "\n"

        for i, row in enumerate(array):
            markdown += (
                f"| {row_header[i]} | "
                if isinstance(row_header[i], str)
                else f"| {row_header[i]:.4e} | "
            )
            markdown += " | ".join(f"{e:.4e}" if not isinstance(e, str) else e for e in row)

            markdown += "|\n"

        return MarkdownStr(markdown)

    def reduced(self, compartments: list[str]) -> np.ndarray:
        """The reduced representation of the KMatrix as numpy array.

        Parameters
        ----------
        compartments :
            The compartment order.
        """

        compartments = [c for c in compartments if c in self.involved_compartments()]
        size = len(compartments)
        array = np.zeros((size, size), dtype=np.float64)
        # Matrix is a dict
        for index in self.matrix:
            i = compartments.index(index[0])
            j = compartments.index(index[1])
            array[i, j] = self.matrix[index]
        return array

    def full(self, compartments: list[str]) -> np.ndarray:
        """The full representation of the KMatrix as numpy array.

        Parameters
        ----------
        compartments :
            The compartment order.
        """
        compartments = [c for c in compartments if c in self.involved_compartments()]
        size = len(compartments)
        mat = np.zeros((size, size), np.float64)
        for (to_comp, from_comp), param in self.matrix.items():
            to_idx = compartments.index(to_comp)
            fr_idx = compartments.index(from_comp)

            if to_idx == fr_idx:
                mat[to_idx, fr_idx] -= param
            else:
                mat[to_idx, fr_idx] += param
                mat[fr_idx, fr_idx] -= param
        return mat

    def eigen(self, compartments: list[str]) -> tuple[np.ndarray, np.ndarray]:
        """Returns the eigenvalues and eigenvectors of the k matrix.

        Parameters
        ----------
        compartments :
            The compartment order.
        """
        # We take the transpose to be consistent with timp
        matrix = self.full(compartments).T
        # get the eigenvectors and values, we take the left ones to have
        # computation consistent with TIMP
        eigenvalues, eigenvectors = eig(matrix, left=True, right=False)
        return (eigenvalues.real, eigenvectors.real)

    def rates(self, initial_concentration: InitialConcentration) -> np.ndarray:
        """The resulting rates of the matrix.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        if self.is_unibranched(initial_concentration):
            return np.diag(self.full(initial_concentration.compartments)).copy()
        rates, _ = self.eigen(initial_concentration.compartments)
        return rates

    def _gamma(
        self,
        eigenvectors: np.ndarray,
        initial_concentration: InitialConcentration,
    ) -> np.ndarray:
        compartments = [
            c for c in initial_concentration.compartments if c in self.involved_compartments()
        ]
        initial_concentration = [
            initial_concentration.parameters[initial_concentration.compartments.index(c)]
            for c in compartments
        ]

        gamma = solve(eigenvectors, initial_concentration)
        return np.diag(gamma)

    def a_matrix(self, initial_concentration: InitialConcentration) -> np.ndarray:
        """The resulting A matrix of the KMatrix.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        return (
            self.a_matrix_unibranch(initial_concentration)
            if self.is_unibranched(initial_concentration)
            else self.a_matrix_non_unibranch(initial_concentration)
        )

    def a_matrix_non_unibranch(self, initial_concentration: InitialConcentration) -> np.ndarray:
        """The resulting A matrix of the KMatrix for a non-unibranched model.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        eigenvalues, eigenvectors = self.eigen(initial_concentration.compartments)
        gamma = self._gamma(eigenvectors, initial_concentration)

        a_matrix = eigenvectors @ gamma

        return a_matrix.T

    def a_matrix_unibranch(self, initial_concentration: InitialConcentration) -> np.ndarray:
        """The resulting A matrix of the KMatrix for an unibranched model.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        compartments = [
            c for c in initial_concentration.compartments if c in self.involved_compartments()
        ]
        matrix = self.full(compartments).T
        rates = np.diag(matrix)

        a_matrix = np.zeros(matrix.shape, dtype=np.float64)
        a_matrix[0, 0] = 1.0
        for i, j in itertools.product(range(rates.size), range(1, rates.size)):
            if i > j:
                continue
            a_matrix[i, j] = np.prod([rates[m] for m in range(j)]) / np.prod(
                [rates[m] - rates[i] for m in range(j + 1) if i != m]
            )

        return a_matrix

    def is_unibranched(self, initial_concentration: InitialConcentration) -> bool:
        """Returns true in the KMatrix represents an unibranched model.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        if (
            np.sum(
                [
                    initial_concentration.parameters[initial_concentration.compartments.index(c)]
                    for c in self.involved_compartments()
                ]
            )
            != 1
        ):
            return False
        matrix = self.reduced(initial_concentration.compartments)
        return not any(
            np.nonzero(matrix[:, i])[0].size != 1 or i != 0 and matrix[i, i - 1] == 0
            for i in range(matrix.shape[1])
        )



"""Functions for calculating conditionally linear parameters and residual with the non-negative
least-squares method."""

import typing

import numpy as np
from scipy.optimize import nnls


def residual_nnls(matrix: np.ndarray, data: np.ndarray) -> typing.Tuple[np.ndarray, np.ndarray]:
    """Calculate the conditionally linear parameters and residual with the nnls method.

    nnls stands for 'non-negative least-squares'.

    Parameters
    ----------
    matrix :
        The model matrix.
    data : np.ndarray
        The data to analyze.
    """
    clp, _ = nnls(matrix, data)
    residual = data - np.dot(matrix, clp)
    return clp, residual



"""Functions for simulating a global analysis model."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from glotaran.analysis.util import calculate_matrix
from glotaran.model import DatasetModel

if TYPE_CHECKING:
    from glotaran.model import Model
    from glotaran.parameter import ParameterGroup


def simulate(
    model: Model,
    dataset: str,
    parameters: ParameterGroup,
    coordinates: dict[str, np.ndarray],
    clp: xr.DataArray | None = None,
    noise: bool = False,
    noise_std_dev: float = 1.0,
    noise_seed: int | None = None,
):
    """Simulates a model.

    Parameters
    ----------
    model :
        The model to simulate.
    parameter :
        The parameters for the simulation.
    dataset :
        Label of the dataset to simulate
    axes :
        A dictionary with axes for simulation.
    clp :
        conditionally linear parameters. Will be used instead of `model.global_matrix` if given.
    noise :
        Add noise to the simulation.
    noise_std_dev :
        The standard deviation for noise simulation.
    noise_seed :
        The seed for the noise simulation.
    """

    dataset_model = model.dataset[dataset].fill(model, parameters)
    dataset_model.set_coordinates(coordinates)

    if dataset_model.has_global_model():
        result = simulate_global_model(
            dataset_model,
            parameters,
            clp,
        )
    elif clp is None:
        raise ValueError(
            f"Cannot simulate dataset {dataset} without global megacomplex " "and no clp provided."
        )
    else:
        result = simulate_clp(
            dataset_model,
            parameters,
            clp,
        )

    if noise:
        if noise_seed is not None:
            np.random.seed(noise_seed)
        result["data"] = (result.data.dims, np.random.normal(result.data, noise_std_dev))

    return result


def simulate_clp(
    dataset_model: DatasetModel,
    parameters: ParameterGroup,
    clp: xr.DataArray,
):

    if "clp_label" not in clp.coords:
        raise ValueError("Missing coordinate 'clp_label' in clp.")
    global_dimension = next(dim for dim in clp.coords if dim != "clp_label")

    global_axis = clp.coords[global_dimension]
    matrices = (
        [
            calculate_matrix(
                dataset_model,
                {global_dimension: index},
            )
            for index, _ in enumerate(global_axis)
        ]
        if dataset_model.is_index_dependent()
        else calculate_matrix(dataset_model, {})
    )

    model_dimension = dataset_model.get_model_dimension()
    model_axis = dataset_model.get_coordinates()[model_dimension]
    result = xr.DataArray(
        data=0.0,
        coords=[
            (model_dimension, model_axis.data),
            (global_dimension, global_axis.data),
        ],
    )
    result = result.to_dataset(name="data")
    for i in range(global_axis.size):
        index_matrix = matrices[i] if dataset_model.is_index_dependent() else matrices
        result.data[:, i] = np.dot(
            index_matrix.matrix,
            clp.isel({global_dimension: i}).sel({"clp_label": index_matrix.clp_labels}),
        )

    return result


def simulate_global_model(
    dataset_model: DatasetModel,
    parameters: ParameterGroup,
    clp: xr.DataArray = None,
):
    """Simulates a global model."""

    # TODO: implement full model clp
    if clp is not None:
        raise NotImplementedError("Simulation of full models with clp is not supported yet.")

    if any(m.index_dependent(dataset_model) for m in dataset_model.global_megacomplex):
        raise ValueError("Index dependent models for global dimension are not supported.")

    global_matrix = calculate_matrix(dataset_model, {}, as_global_model=True)
    global_clp_labels = global_matrix.clp_labels
    global_matrix = xr.DataArray(
        global_matrix.matrix.T,
        coords=[
            ("clp_label", global_clp_labels),
            (dataset_model.get_global_dimension(), dataset_model.get_global_axis()),
        ],
    )

    return simulate_clp(
        dataset_model,
        parameters,
        global_matrix,
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import numba as nb
import numpy as np
import xarray as xr

from glotaran.builtin.megacomplexes.decay.irf import IrfMultiGaussian
from glotaran.builtin.megacomplexes.decay.irf import IrfSpectralMultiGaussian
from glotaran.model import DatasetModel

if TYPE_CHECKING:
    from glotaran.builtin.megacomplexes.decay.decay_megacomplex import DecayMegacomplex


def decay_matrix_implementation(
    matrix: np.ndarray,
    rates: np.ndarray,
    global_index: int,
    global_axis: np.ndarray,
    model_axis: np.ndarray,
    dataset_model: DatasetModel,
):
    if isinstance(dataset_model.irf, IrfMultiGaussian):

        (
            centers,
            widths,
            irf_scales,
            shift,
            backsweep,
            backsweep_period,
        ) = dataset_model.irf.parameter(global_index, global_axis)

        for center, width, irf_scale in zip(centers, widths, irf_scales):
            calculate_decay_matrix_gaussian_irf(
                matrix,
                rates,
                model_axis,
                center - shift,
                width,
                irf_scale,
                backsweep,
                backsweep_period,
            )
        if dataset_model.irf.normalize:
            matrix /= np.sum(irf_scale)

    else:
        calculate_decay_matrix_no_irf(matrix, rates, model_axis)


@nb.jit(nopython=True, parallel=True)
def calculate_decay_matrix_no_irf(matrix, rates, times):
    for n_r in nb.prange(rates.size):
        r_n = rates[n_r]
        for n_t in range(times.size):
            t_n = times[n_t]
            matrix[n_t, n_r] += np.exp(r_n * t_n)


sqrt2 = np.sqrt(2)


@nb.jit(nopython=True, parallel=True)
def calculate_decay_matrix_gaussian_irf(
    matrix, rates, times, center, width, scale, backsweep, backsweep_period
):
    """Calculates a decay matrix with a gaussian irf."""
    for n_r in nb.prange(rates.size):
        r_n = -rates[n_r]
        backsweep_valid = abs(r_n) * backsweep_period > 0.001
        alpha = (r_n * width) / sqrt2
        for n_t in nb.prange(times.size):
            t_n = times[n_t]
            beta = (t_n - center) / (width * sqrt2)
            thresh = beta - alpha
            if thresh < -1:
                matrix[n_t, n_r] += scale * 0.5 * erfcx(-thresh) * np.exp(-beta * beta)
            else:
                matrix[n_t, n_r] += (
                    scale * 0.5 * (1 + erf(thresh)) * np.exp(alpha * (alpha - 2 * beta))
                )
            if backsweep and backsweep_valid:
                x1 = np.exp(-r_n * (t_n - center + backsweep_period))
                x2 = np.exp(-r_n * ((backsweep_period / 2) - (t_n - center)))
                x3 = np.exp(-r_n * backsweep_period)
                matrix[n_t, n_r] += scale * (x1 + x2) / (1 - x3)


import ctypes  # noqa: E402

# This is a work around to use scipy.special function with numba
from numba.extending import get_cython_function_address  # noqa: E402

_dble = ctypes.c_double

functype = ctypes.CFUNCTYPE(_dble, _dble)

erf_addr = get_cython_function_address("scipy.special.cython_special", "__pyx_fuse_1erf")
erfcx_addr = get_cython_function_address("scipy.special.cython_special", "__pyx_fuse_1erfcx")

erf = functype(erf_addr)
erfcx = functype(erfcx_addr)


def retrieve_species_associated_data(
    dataset_model: DatasetModel,
    dataset: xr.Dataset,
    species_dimension: str,
    global_dimension: str,
    name: str,
    is_full_model: bool,
    as_global: bool,
):
    species = dataset_model.initial_concentration.compartments
    model_dimension = dataset_model.get_model_dimension()
    if as_global:
        model_dimension, global_dimension = global_dimension, model_dimension
    dataset.coords[species_dimension] = species
    matrix = dataset.global_matrix if as_global else dataset.matrix
    clp_dim = "global_clp_label" if as_global else "clp_label"

    if len(dataset.matrix.shape) == 3:
        #  index dependent
        dataset["species_concentration"] = (
            (
                global_dimension,
                model_dimension,
                species_dimension,
            ),
            matrix.sel({clp_dim: species}).values,
        )
    else:
        #  index independent
        dataset["species_concentration"] = (
            (
                model_dimension,
                species_dimension,
            ),
            matrix.sel({clp_dim: species}).values,
        )

    if not is_full_model:
        dataset[f"species_associated_{name}"] = (
            (
                global_dimension,
                species_dimension,
            ),
            dataset.clp.sel(clp_label=species).data,
        )


def retrieve_decay_associated_data(
    megacomplex: DecayMegacomplex,
    dataset_model: DatasetModel,
    dataset: xr.Dataset,
    global_dimension: str,
    name: str,
    multiple_complexes: bool,
):
    k_matrix = megacomplex.full_k_matrix()

    species = dataset_model.initial_concentration.compartments
    species = [c for c in species if c in k_matrix.involved_compartments()]

    matrix = k_matrix.full(species)
    matrix_reduced = k_matrix.reduced(species)
    a_matrix = k_matrix.a_matrix(dataset_model.initial_concentration)
    rates = k_matrix.rates(dataset_model.initial_concentration)
    lifetimes = 1 / rates

    das = dataset[f"species_associated_{name}"].sel(species=species).values @ a_matrix.T

    component_coords = {
        "component": np.arange(rates.size),
        "rate": ("component", rates),
        "lifetime": ("component", lifetimes),
    }
    das_coords = component_coords.copy()
    das_coords[global_dimension] = dataset.coords[global_dimension]
    das_name = f"decay_associated_{name}"
    das = xr.DataArray(das, dims=(global_dimension, "component"), coords=das_coords)

    a_matrix_coords = component_coords.copy()
    a_matrix_coords["species"] = species
    a_matrix_name = "a_matrix"
    a_matrix = xr.DataArray(a_matrix, coords=a_matrix_coords, dims=("component", "species"))

    k_matrix_name = "k_matrix"
    k_matrix = xr.DataArray(matrix, coords=[("to_species", species), ("from_species", species)])

    k_matrix_reduced_name = "k_matrix_reduced"
    k_matrix_reduced = xr.DataArray(
        matrix_reduced, coords=[("to_species", species), ("from_species", species)]
    )

    if multiple_complexes:
        das_name = f"decay_associated_{name}_{megacomplex.label}"
        das = das.rename(component=f"component_{megacomplex.label}")
        a_matrix_name = f"a_matrix_{megacomplex.label}"
        a_matrix = a_matrix.rename(component=f"component_{megacomplex.label}")
        k_matrix_name = f"k_matrix_{megacomplex.label}"
        k_matrix_reduced_name = f"k_matrix_reduced_{megacomplex.label}"

    dataset[das_name] = das
    dataset[a_matrix_name] = a_matrix
    dataset[k_matrix_name] = k_matrix
    dataset[k_matrix_reduced_name] = k_matrix_reduced


def retrieve_irf(dataset_model: DatasetModel, dataset: xr.Dataset, global_dimension: str):

    irf = dataset_model.irf
    model_dimension = dataset_model.get_model_dimension()

    dataset["irf"] = (
        (model_dimension),
        irf.calculate(
            index=0,
            global_axis=dataset.coords[global_dimension].values,
            model_axis=dataset.coords[model_dimension].values,
        ).data,
    )
    center = irf.center if isinstance(irf.center, list) else [irf.center]
    width = irf.width if isinstance(irf.width, list) else [irf.width]
    dataset["irf_center"] = ("irf_nr", center) if len(center) > 1 else center[0]
    dataset["irf_width"] = ("irf_nr", width) if len(width) > 1 else width[0]
    if isinstance(irf, IrfSpectralMultiGaussian) and irf.dispersion_center:
        dataset["irf_center_location"] = (
            ("irf_nr", global_dimension),
            irf.calculate_dispersion(dataset.coords["spectral"].values),
        )
        # center_dispersion_1 for backwards compatibility (0.3-0.4.1)
        dataset["center_dispersion_1"] = dataset["irf_center_location"].sel(irf_nr=0)



import numpy as np

from glotaran.analysis.util import calculate_matrix
from glotaran.builtin.megacomplexes.baseline import BaselineMegacomplex
from glotaran.builtin.megacomplexes.decay import DecayMegacomplex
from glotaran.model import Model
from glotaran.parameter import ParameterGroup


def test_baseline():
    model = Model.from_dict(
        {
            "initial_concentration": {
                "j1": {"compartments": ["s1"], "parameters": ["2"]},
            },
            "megacomplex": {
                "mc1": {"type": "decay", "k_matrix": ["k1"]},
                "mc2": {"type": "baseline", "dimension": "time"},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s1", "s1"): "1",
                    }
                }
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "megacomplex": ["mc1", "mc2"],
                },
            },
        },
        megacomplex_types={"decay": DecayMegacomplex, "baseline": BaselineMegacomplex},
    )

    parameter = ParameterGroup.from_list(
        [
            101e-4,
            [1, {"vary": False, "non-negative": False}],
            [42, {"vary": False, "non-negative": False}],
        ]
    )

    time = np.asarray(np.arange(0, 50, 1.5))
    pixel = np.asarray([0])
    coords = {"time": time, "pixel": pixel}
    dataset_model = model.dataset["dataset1"].fill(model, parameter)
    dataset_model.overwrite_global_dimension("pixel")
    dataset_model.set_coordinates(coords)
    matrix = calculate_matrix(dataset_model, {})
    compartments = matrix.clp_labels

    assert len(compartments) == 2
    assert "dataset1_baseline" in compartments

    assert matrix.matrix.shape == (time.size, 2)
    assert np.all(matrix.matrix[:, 1] == 1)



"""Helper functions."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any
    from typing import Callable
    from typing import TypeVar

    DecoratedFunc = TypeVar("DecoratedFunc", bound=Callable[..., Any])  # decorated function


class ModelError(Exception):
    """Raised when a model contains errors."""

    def __init__(self, error: str):
        super().__init__(f"ModelError: {error}")


def wrap_func_as_method(
    cls: Any, name: str = None, annotations: dict[str, type] = None, doc: str = None
) -> Callable[[DecoratedFunc], DecoratedFunc]:
    """A decorator to wrap a function as class method.

    Notes
    -----

    Only for internal use.

    Parameters
    ----------
    cls :
        The class in which the function will be wrapped.
    name :
        The name of method. If `None`, the original function's name is used.
    annotations :
        The annotations of the method. If `None`, the original function's annotations are used.
    doc :
        The documentation of the method. If `None`, the original function's documentation is used.
    """

    def wrapper(func: DecoratedFunc) -> DecoratedFunc:
        if name:
            func.__name__ = name
        if annotations:
            setattr(func, "__annotations__", annotations)
        if doc:
            func.__doc__ = doc
        func.__qualname__ = cls.__qualname__ + "." + func.__name__
        func.__module__ = cls.__module__

        return func

    return wrapper
